import unittest
import os
import tempfile


from fuse.data.datasets.dataset_default import DatasetDefault
from fuse.data.datasets.caching.samples_cacher import SamplesCacher

from fuseimg.datasets.kits21 import KITS21


class TestPipelineCaching(unittest.TestCase):
    def test_basic_1(self):
        """
        Test basic imaging ops
        """
        tmpdir = tempfile.mkdtemp()
        kits_dir = os.path.join(tmpdir, "kits21")
        cases = [100, 150, 200]
        KITS21.download(kits_dir, cases)

        static_pipeline = KITS21.static_pipeline(kits_dir)
        dynamic_pipeline = KITS21.dynamic_pipeline()

        cache_dirs = [
            os.path.join(tmpdir, "cache_a"),
            os.path.join(tmpdir, "cache_b"),
        ]

        cacher = SamplesCacher("fuseimg_ops_testing_cache", static_pipeline, cache_dirs)

        sample_ids = [f"case_{_:05}" for _ in cases]
        ds = DatasetDefault(
            sample_ids,
            static_pipeline,
            dynamic_pipeline=dynamic_pipeline,
            cacher=cacher,
        )

        ds.create()


if __name__ == "__main__":
    unittest.main()



import unittest

from fuse.utils.cpu_profiling import Timer


class TestTimer(unittest.TestCase):
    def test_timer_1(self):
        """ """

        def foo(x):
            with Timer("banana"):
                for _ in range(10**5):
                    x = x * 2

        foo(20)


if __name__ == "__main__":
    unittest.main()



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""
from typing import Optional, Sequence
import pandas as pds

from fuse.data.datasets.dataset_base import DatasetBase

from fuse.utils.file_io.file_io import save_dataframe


class ExportDataset:
    """
    Export data
    """

    @staticmethod
    def export_to_dataframe(
        dataset: DatasetBase,
        keys: Sequence[str],
        output_filename: Optional[str] = None,
        sample_id_key: str = "data.sample_id",
        **dataset_get_kwargs
    ) -> pds.DataFrame:
        """
        extract from dataset the specified and keys and create a dataframe.
        If output_filename will be specified, the dataframe will also be saved in a file.
        :param dataset: the dataset to extract the values from
        :param keys: keys to extract from sample_dict
        :param output_filename: Optional, if set, will save the dataframe into a file.
                                The file type will be inferred from filename, see fuse.utils.file_io.file_io.save_dataframe for more details
        :param dataset_get_kwargs: additional parameters to dataset.get(), might be used to optimize the running time
        """
        # add sample_id to keys list
        if keys is not None:
            all_keys = []
            all_keys += list(keys)
            if sample_id_key not in keys:
                all_keys.append(sample_id_key)
        else:
            all_keys = None

        # read all the data
        data = dataset.get_multi(keys=all_keys, **dataset_get_kwargs)

        # store in dataframe
        df = pds.DataFrame()

        for key in all_keys:
            values = [sample_dict[key] for sample_dict in data]
            df[key] = values

        if output_filename is not None:
            save_dataframe(df, output_filename)

        return df



from .backbone_inception_resnet_v2 import BackboneInceptionResnetV2
from .backbone_mlp import BackboneMultilayerPerceptron
from .backbone_resnet_3d import BackboneResnet3D
from .backbone_resnet import BackboneResnet



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Tuple, Any

import torch.nn as nn
from torch import Tensor
from torch.hub import load_state_dict_from_url
from torchvision.models.video.resnet import VideoResNet, BasicBlock, Conv3DSimple, BasicStem, model_urls


class BackboneResnet3D(VideoResNet):
    """
    3D model classifier (ResNet architecture"
    """

    def __init__(self, pretrained: bool = False, in_channels: int = 3, name: str = "r3d_18") -> None:
        """
        Create 3D ResNet model
        :param pretrained: Use pretrained weights
        :param in_channels: number of input channels
        :param name: model name. currently only 'r3d_18' is supported
        """
        # init parameters per required backbone
        init_parameters = {
            "r3d_18": {
                "block": BasicBlock,
                "conv_makers": [Conv3DSimple] * 4,
                "layers": [2, 2, 2, 2],
                "stem": BasicStem,
            },
        }[name]
        # init original model
        super().__init__(**init_parameters)

        # load pretrained parameters if required
        if pretrained:
            state_dict = load_state_dict_from_url(model_urls[name])
            self.load_state_dict(state_dict)

        # save input parameters
        self.pretrained = pretrained
        self.in_channels = in_channels

        if self.in_channels != 3:
            # override the first convolution layer to support any number of input channels
            self.stem = nn.Sequential(
                nn.Conv3d(self.in_channels, 64, kernel_size=(3, 7, 7), stride=(1, 2, 2), padding=(1, 3, 3), bias=False),
                nn.BatchNorm3d(64),
                nn.ReLU(inplace=True),
            )

    def features(self, x: Tensor) -> Any:
        """
        Extract spatial features - given a 3D tensor
        :param x: Input tensor - shape: [batch_size, channels, z, y, x]
        :return: spatial features - shape [batch_size, n_features, z', y', x']
        """
        x = self.stem(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        return x

    def forward(self, x: Tensor) -> Tuple[Tensor, None, None, None]:  # type: ignore
        """
        Forward pass. 3D global classification given a volume
        :param x: Input volume. shape: [batch_size, channels, z, y, x]
        :return: logits for global classification. shape: [batch_size, n_classes].
        """
        x = self.features(x)
        return x



import math
import numbers
from typing import Any, Callable, Dict, List, Optional, OrderedDict, Sequence, Tuple, Union
from fuse.data.key_types import TypeDetectorBase
import copy
from enum import Enum
from .op_base import OpBase, OpReversibleBase, op_call, op_reverse  # DataType,
from fuse.data.patterns import Patterns
from fuse.utils.ndict import NDict
import numpy as np


class OpRepeat(OpReversibleBase):
    """
    Repeat an op multiple times

    Typically used to apply the same operation on a list of keys in sample_dict
    Example:
    "

    repeat_for =

        #...
        (OpRepeat(OpCropToMinimalBBox(),
            [dict(key='data.cc.image'), dict(key='data.mlo.image'),dict(key='data.mlo.seg', margin=100)] #per provided dict a new OpCropToMinimalBBox invocation will be triggered
            )),
            dict(margin=12)), #this value will be passed to all OpCropToMinimalBBox invocations
        #...
    ]

    note - the values in provided in the list of dicts will *override* any kwargs
    In the example above, margin=12 will be used for both 'data.cc.image' and 'data.mlo.image',
        but a value of margin=100 will be used for 'data.mlo.seg'

    "
    """

    def __init__(self, op: OpBase, kwargs_per_step_to_add: Sequence[dict]):
        """
        See example above
        :param op: the operation to repeat
        :param kwargs_per_step_to_add: sequence of arguments (kwargs format) specific for a single repetition. those arguments will be added/overide the kwargs provided in __call__() function.
        """
        super().__init__()
        self._op = op
        self._kwargs_per_step_to_add = kwargs_per_step_to_add

    def __call__(self, sample_dict: NDict, op_id: Optional[str], **kwargs) -> Union[None, dict, List[dict]]:
        """
        See super class
        """

        for step_index, step_kwargs_to_add in enumerate(self._kwargs_per_step_to_add):
            step_kwargs = copy.copy(kwargs)
            step_kwargs.update(step_kwargs_to_add)
            full_step_id = f"{op_id}_{step_index}"
            sample_dict[full_step_id + "_debug_info.op_name"] = self._op.__class__.__name__
            sample_dict = op_call(self._op, sample_dict, full_step_id, **step_kwargs)

            assert not isinstance(
                sample_dict, list
            ), f"splitting samples within {type(self).__name__} operation is not supported"

            if sample_dict is None:
                return None
            elif not isinstance(sample_dict, dict):
                raise Exception(f"unexpected sample_dict type {type(sample_dict)}")

        return sample_dict

    def reverse(self, sample_dict: NDict, key_to_reverse: str, key_to_follow: str, op_id: Optional[str]) -> dict:
        """
        See super class
        """
        for step_index in reversed(range(len(self._kwargs_per_step_to_add))):
            sample_dict = op_reverse(self._op, sample_dict, key_to_reverse, key_to_follow, f"{op_id}_{step_index}")

        return sample_dict


class OpLambda(OpReversibleBase):
    """
    Apply simple lambda function / function to transform single value from sample_dict (or the all dictionary)
    Optionally add reverse method if required.
    Example:
    OpLambda(func=lambda x: torch.tensor(x))
    """

    def __init__(self, func: Callable, func_reverse: Optional[Callable] = None, **kwargs):
        super().__init__(**kwargs)
        self._func = func
        self._func_reverse = func_reverse

    def __call__(
        self, sample_dict: NDict, op_id: Optional[str], key: Optional[str] = None, **kwargs
    ) -> Union[None, dict, List[dict]]:
        """
        More details in super class
        :param key: apply lambda func on sample_dict[key]. If none the input and output of the lambda function are the entire sample_dict
        """
        sample_dict[op_id] = key
        if key is not None:
            value = sample_dict[key]
            value = self._func(value, **kwargs)
            sample_dict[key] = value
        else:
            sample_dict = self._func(sample_dict)

        return sample_dict

    def reverse(self, sample_dict: NDict, key_to_reverse: str, key_to_follow: str, op_id: Optional[str]) -> dict:
        """
        See super class
        """
        key = sample_dict[op_id]
        if key is not None:
            if key == key_to_follow:
                value = sample_dict[key_to_reverse]
                value = self._func_reverse(value)
                sample_dict[key_to_reverse] = value
        else:
            sample_dict = self._func_reverse(sample_dict)

        return sample_dict


class OpFunc(OpReversibleBase):
    """
    Helps to wrap an existing simple python function without writing boilerplate code.

    The wrapped function format is:

    def foo(*, *kwargs) -> Tuple:
        pass


    Example:

    def add_seperator(text:str, sep=' '):
        return sep.join(text)

    OpAddSeperator = OpFunc(add_seperator)

    usage in pipeline:

    pipeline = [
        (OpAddSeperator, dict(inputs={'data.text_input':'text'}, outputs='data.text_input'), #
    ]


    """

    def __init__(self, func: Callable, **kwargs):
        """
        :param func: a callable to call in  __call__()
        """
        super().__init__(**kwargs)
        self._func = func

    def __call__(
        self,
        sample_dict: NDict,
        op_id: Optional[str],
        inputs: Dict[str, str],
        outputs: Union[Sequence[str], str],
        **kwargs,
    ) -> Union[None, dict, List[dict]]:
        """
        See super class
        :param inputs: dictionary that map between the key_name of a value stored in sample_dict the the input argument name in func
        :param outputs: sequence of key_names to store each return value of func.

        """
        # extract inputs from sample dict
        kwargs_from_sample_dict = {}
        for input_key_name, func_arg_name in inputs.items():
            value = sample_dict[input_key_name]
            kwargs_from_sample_dict[func_arg_name] = value

        # all kwargs
        all_kwargs = copy.copy(kwargs)
        all_kwargs.update(kwargs_from_sample_dict)
        func_outputs = self._func(**all_kwargs)

        # add to sample_dict
        if isinstance(outputs, str):
            sample_dict[outputs] = func_outputs
        elif isinstance(outputs, Sequence):
            assert len(func_outputs) == len(
                outputs
            ), f"expecting that function {self._func} will output {len(outputs)} values"
            for output_name, output_value in zip(outputs, func_outputs):
                sample_dict[output_name] = output_value
        else:
            raise Exception(
                f"expecting outputs to be either str or sequence of str. got {type(self._outputs).__name__}"
            )

        return sample_dict


class OpApplyPatterns(OpReversibleBase):
    """
    Select and apply an operation according to key name.
    Instead of specifying every relevant key, the op will be applied for every key that matched a specified pattern
    Example:
    patterns_dict = OrderedDict([(r"^.*.cc.img$|^.*.cc.seg$", (op_affine, dict(rotate=Uniform(-90.0, 90.0))),
                                (r"^.*.mlo.img$|^.*.mlo.seg$", (op_affine, dict(rotate=Uniform(-45.0, 54.0)))])
    op_apply_pat = OpApplyPatterns(patterns_dict)
    """

    def __init__(self, patterns_dict: Optional[OrderedDict] = None):
        """
        :param patterns_dict: map a regex pattern to a pair of op and arguments (will be added/override the arguments provided in __call__() function).
                             For given value in a sample dict, it will look for the first match in the order dict and will apply the op on this specific key.
                             The ops specified in patterns_dict, must implement a __call__ method with an argument called key.
        """
        super().__init__()
        self._patterns_dict = Patterns(patterns_dict, (None, None))

    def __call__(self, sample_dict: NDict, op_id: Optional[str], **kwargs) -> Union[None, dict, List[dict]]:
        """
        See super class
        """

        for key in sample_dict.keypaths():
            op, op_kwargs_to_add = self._patterns_dict.get_value(key)
            if op is None:
                continue

            op_kwargs = copy.copy(kwargs)
            op_kwargs.update(op_kwargs_to_add)
            sample_dict = op_call(op, sample_dict, f"{op_id}_{key}", key=key, **op_kwargs)

            assert not isinstance(
                sample_dict, list
            ), f"splitting samples within {type(self).__name__} operation is not supported"

            if sample_dict is None:
                return None
            elif not isinstance(sample_dict, dict):
                raise Exception(f"unexpected sample_dict type {type(sample_dict)}")

        return sample_dict

    def reverse(self, sample_dict: NDict, key_to_reverse: str, key_to_follow: str, op_id: Optional[str]) -> dict:
        """
        See super class
        """
        op, _ = self._patterns_dict.get_value(key_to_follow)
        if op is None:
            return

        sample_dict = op_reverse(op, sample_dict, key_to_reverse, key_to_follow, f"{op_id}_{key_to_follow}")

        return sample_dict


class OpApplyTypes(OpReversibleBase):
    """
    Select and apply an operation according value type (inferred from key name). See OpBase for more information about how it is inferred.
    Instead of specifying every relevant key, the op will be applied for every key that matched a specified pattern
    Example:
    types_dict = {  DataType.Image: (op_affine_image, dict()),
                    DataType.Seg: (op_affine_image, dict()),
                     BBox: (op_affine_bbox, dict())}

    op_apply_type = OpApplyTypes(types_dict)
    """

    def __init__(self, type_to_op_dict: Dict[Enum, Tuple[OpBase, dict]], type_detector: TypeDetectorBase):
        """
        :param type_to_op_dict: map a type (See enum DataType) to a pair of op and correspending arguments (will be added/override the arguments provided in __call__() function)
        """
        super().__init__()
        self._type_to_op_dict = type_to_op_dict
        self._type_detector = type_detector

    def __call__(self, sample_dict: NDict, op_id: Optional[str], **kwargs) -> Union[None, dict, List[dict]]:
        """
        See super class
        """
        all_keys = sample_dict.keypaths()
        for key in all_keys:
            key_type = self._type_detector.get_type(sample_dict, key)

            op, op_kwargs_to_add = self._type_to_op_dict.get(key_type, (None, None))
            if op is None:
                continue

            op_kwargs = copy.copy(kwargs)
            op_kwargs.update(op_kwargs_to_add)
            if "key" in op_kwargs:
                raise Exception(
                    'OpApplyTypes::"key" is already found in kwargs. Are you calling OpApplyTypes from within OpApplyTypes? it is not supported.'
                )
            sample_dict = op_call(op, sample_dict, f"{op_id}_{key}", key=key, **op_kwargs)

            assert not isinstance(
                sample_dict, list
            ), f"splitting samples within {type(self).__name__} operation is not supported"

            if sample_dict is None:
                return None
            elif not isinstance(sample_dict, dict):
                raise Exception(f"unexpected sample_dict type {type(sample_dict)}")

        return sample_dict

    def reverse(self, sample_dict: NDict, key_to_reverse: str, key_to_follow: str, op_id: Optional[str]) -> dict:
        """
        See super class
        """
        key_type = self._type_detector.get_type(sample_dict, key_to_follow)
        op, _ = self._type_to_op_dict.get(key_type, (None, None))
        if op is None:
            return

        sample_dict = op_reverse(op, sample_dict, key_to_reverse, key_to_follow, f"{op_id}_{key_to_follow}")

        return sample_dict


class OpCollectMarker(OpReversibleBase):
    """
    Use this op within the dynamic pipeline to optimize the reading time for components such as sampler, export and stats that don't need to read the entire sample.
    OpCollectMarker will specify the last op to call to get all the required information from sample.
    In addition, to avoid from reading the entire sample including images, OpCollectMarker can also specify the list of keys required for the relevant part of the dynamic pipeline.

    Examples:
    1.
    The static pipeline generates a sample including an image ('data.image') and a label ('data.label').
    The training set sampler configured to balance a batch according to 'data.label'
    To optimize the reading time of the sampler:
    Add at the beginning of the dynamic pipeline -
    OpCollectMarker(name="sampler", static_keys_deps=["data.label"])
    2.
    The static pipeline generate an image ('data.image') and a metadata ('data.metadata').
    The dynamic pipeline includes few operations reading 'data.metadata' and that set a value used to balance the class (op_do and op_convert).
    To optimize the reading time of the sampler:
    Move op_do and op_convert to the beginning of the pipeline.
    Add just after them the following op:
    OpCollectMarker(name="sampler", static_kets_deps=["data.metadata"])

    In both cases the sampler can now read subset of the sample using: dataset.get_multi(collect_marker_name="sampler", ..)
    """

    def __init__(self, name: str, static_key_deps: Sequence[str]):
        super().__init__()
        self._name = name
        self._static_keys_deps = static_key_deps

    def get_info(self) -> dict:
        """
        Returns collect marker info including name and static_keys_deps
        """
        return {"name": self._name, "static_keys_deps": self._static_keys_deps}

    def __call__(self, sample_dict: dict, op_id: Optional[str], **kwargs) -> Union[None, dict, List[dict]]:
        return sample_dict

    def reverse(self, sample_dict: dict, key_to_reverse: str, key_to_follow: str, op_id: Optional[str]) -> dict:
        return sample_dict


class OpKeepKeypaths(OpBase):
    """
    Use this op to keep only the defined keypaths in the sample
    A case where this is useful is if you want to limit the amount of data that gets transferred by multiprocessing by DataLoader workers.
    You can keep only what you want to enter the collate.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __call__(self, sample_dict: NDict, keep_keypaths: List[str]) -> Union[None, dict, List[dict]]:
        prev_sample_dict = sample_dict
        sample_dict = NDict()
        for k in keep_keypaths:
            sample_dict[k] = prev_sample_dict[k]

        return sample_dict


class OpLookup(OpBase):
    """
    Convert a value to another value. It should be specified in a dictionary mapping old value to a new value
    Example:
    To read the gender represented by strings "male" and "female" and convert it to int do the following
    (OpLookup(map={"male": 0, "female": 1}). dict(key_in="data.input.gender", key_out="data.input,gender"))
    """

    def __init__(self, map: dict, not_exist_error: bool = True):
        """
        :param not_exist_error: false iff if the value does not exist it will keep the previous value
        """
        super().__init__()
        self._map = map
        self._not_exist_error = not_exist_error

    def __call__(self, sample_dict: NDict, key_in: str, key_out: str) -> Union[None, dict, List[dict]]:
        """
        :param key_in: key to a value
        :param key_out: key to store the converted vale
        """
        value = sample_dict[key_in]
        if value in self._map:
            sample_dict[key_out] = self._map[value]
        elif self._not_exist_error:
            raise Exception(f"value {value} does not exist in mapping")

        return sample_dict


class OpToOneHot(OpBase):
    """
    Map category value to one hot vector
    """

    def __init__(self, num_classes: int):
        """
        :param num_classes: the size of the one hot vector
        """
        super().__init__()
        self._num_classes = num_classes

    def __call__(self, sample_dict: NDict, key_in: str, key_out: str) -> Union[None, dict, List[dict]]:
        """
        :param key_in: key to a class number (int)
        :param key_out: key to store the one hot vector
        """
        value = sample_dict[key_in]
        one_hot = np.zeros(self._num_classes)
        one_hot[value] = 1.0
        sample_dict[key_out] = one_hot

        return sample_dict


class OpConcat(OpBase):
    """
    Concatenate list of numpy arrays along a given axis.
    To create clinical vector that includes all the clinical information about a patient and save into "data.input.clinical" do:
    (OpConcat(), dict(keys_int=["data.input.age", "data.input.gender_one_hot", "data.input.smoking_history"], key_out="data.input_clinical", axis=0)
    """

    def __call__(
        self, sample_dict: NDict, keys_in: Sequence[str], key_out: str, axis: int = 0
    ) -> Union[None, dict, List[dict]]:
        """
        :param keys_in: sequence of keys to numpy arrays we want to concatenate
        :param key_out: the key to store the concatenated vector
        :param axis: concatenate along the specified axis
        """
        values = [np.asarray(sample_dict[key_in]) for key_in in keys_in]
        values = [v if len(v.shape) > 0 else np.expand_dims(v, axis=0) for v in values]
        sample_dict[key_out] = np.concatenate(values, axis=axis)

        return sample_dict


class OpOverrideNaN(OpBase):
    """
    Override missing values (value equals to nan)
    """

    def __call__(self, sample_dict: NDict, key: str, value_to_fill: Any) -> NDict:
        assert key in sample_dict, f"Error: missing {key}, available keys {sample_dict.keypaths()} "
        if isinstance(sample_dict[key], numbers.Number) and math.isnan(sample_dict[key]):
            sample_dict[key] = value_to_fill
        return sample_dict


class OpZScoreNorm(OpBase):
    def __call__(self, sample_dict: NDict, key: str, mean: float, std: float):
        sample_dict[key] = (sample_dict[key] - mean) / std
        return sample_dict



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Optional, Sequence

from .metrics_classification_common import MetricMultiClassDefault
from fuse.eval.metrics.libs.model_comparison import ModelComparison
from fuse.eval.metrics.metrics_common import MetricDefault


class MetricDelongsTest(MetricMultiClassDefault):
    def __init__(self, pred1: str, pred2: str, target: str, class_names: Optional[Sequence[str]] = None, **kwargs):
        # :param pred1: key name for the predictions of model 1
        # :param pred2: key name for the predictions of model 2
        # :param target: key name for the ground truth labels
        # :param class_names: class names. required for multi-class classifiers

        super().__init__(
            pred=None,
            target=target,
            metric_func=ModelComparison.delong_auc_test,
            class_names=class_names,
            pred1=pred1,
            pred2=pred2,
            **kwargs
        )


class MetricContingencyTable(MetricDefault):
    def __init__(self, var1: str, var2: str, **kwargs):
        """
        Create contingency table from two paired variables.
        :param var1: key name for the first variable
        :param var2: key name for the second variable
        """
        super().__init__(
            pred=None, target=None, metric_func=ModelComparison.contingency_table, var1=var1, var2=var2, **kwargs
        )


class MetricMcnemarsTest(MetricMultiClassDefault):
    def __init__(self, pred1: str, pred2: str, target: Optional[str] = None, exact: Optional[bool] = True, **kwargs):
        """
        McNemar's statistical test for comparing two model's predictions or accuracies
        in the sense of the statistics of their disagreements, as seen in the contingency table.
        For comparing two classifiers, it's possible to provide the ground truth and then the comparison
        will be of the models' accuracies. It's also possible to perform significance test on the predictions only.
        This can be useful either for classifiers, or any other paired data, even in cases where ground truth
        may not be applicable/available (i.e patients' opinion surveys, unsupervised model predictions, etc'.).
        :param pred1: key name for the class predictions of the first model
        :param pred2: key name for the class predictions of the second model
        :param target: (optional) key name for the ground truth
        :param exact: If True, then the binomial distribution will be used. Otherwise, the chi-square distribution, which is the approximation to the distribution of the test statistic for large sample sizes.
            The exact test is recommended for small (<25) number of discordants in the contingency table
        """
        super().__init__(
            pred=None,
            pred1=pred1,
            pred2=pred2,
            target=target,
            exact=exact,
            metric_func=ModelComparison.mcnemars_test,
            **kwargs
        )



from fuse.data.datasets.dataset_default import DatasetDefault
from fuse.data.datasets.caching.samples_cacher import SamplesCacher
from fuseimg.data.ops.image_loader import OpLoadImage
from fuseimg.data.ops.color import OpNormalizeAgainstSelf
from fuseimg.data.ops.shape_ops import OpFlipBrightSideOnLeft2D, OpFindBiggestNonEmptyBbox2D, OpResizeAndPad2D
from fuse.data import PipelineDefault, OpToTensor
from fuse.data.ops.ops_common import OpLambda
from fuseimg.data.ops.aug.color import OpAugColor
from fuseimg.data.ops.aug.geometry import OpAugAffine2D
from fuse.data.ops.ops_aug_common import OpSample, OpRandApply
from fuse.data.ops.ops_read import OpReadDataframe
from fuse.data.ops.ops_cast import OpToNumpy
from fuse.data.ops.op_base import OpBase
from fuse.utils import NDict
from functools import partial
from typing import Hashable, Optional, Sequence
import torch
import pandas as pd
import numpy as np
import pydicom
import os
import glob
from pathlib import Path
from fuse.data.utils.sample import get_sample_id
from fuse.utils.rand.param_sampler import Uniform, RandInt, RandBool


class OpCMMDSampleIDDecode(OpBase):
    """
    decodes sample id into image and segmentation filename
    """

    def __call__(self, sample_dict: NDict) -> NDict:
        """ """
        sid = get_sample_id(sample_dict)

        img_filename_key = "data.input.img_path"
        sample_dict[img_filename_key] = sid

        return sample_dict


class CMMD:
    """
    # dataset that contains breast mamography biopsy info and metadata from chinese patients
    # Path to the stored dataset location
    # dataset should be download from https://wiki.cancerimagingarchive.net/pages/viewpage.action?pageId=70230508
    # download requires NBIA data retriever https://wiki.cancerimagingarchive.net/display/NBIA/Downloading+TCIA+Images
    # put on the folliwing in the main folder  -
    # 1. CMMD_clinicaldata_revision.csv which is a converted version of CMMD_clinicaldata_revision.xlsx
    # 2. folder named CMMD which is the downloaded data folder
    """

    # bump whenever the static pipeline modified
    CMMD_DATASET_VER = 0

    @staticmethod
    def static_pipeline(data_dir: str, data_source: pd.DataFrame, target: str) -> PipelineDefault:
        """
        Get suggested static pipeline (which will be cached), typically loading the data plus design choices that we won't experiment with.
        :param data_path: path to original kits21 data (can be downloaded by KITS21.download())
        """
        static_pipeline = PipelineDefault(
            "cmmd_static",
            [
                # decoding sample ID
                (OpCMMDSampleIDDecode(), dict()),  # will save image and seg path to "data.input.img_path"
                (OpLoadImage(data_dir), dict(key_in="data.input.img_path", key_out="data.input.img", format="dcm")),
                (OpFlipBrightSideOnLeft2D(), dict(key="data.input.img")),
                (OpFindBiggestNonEmptyBbox2D(), dict(key="data.input.img")),
                (OpNormalizeAgainstSelf(), dict(key="data.input.img")),
                (OpToNumpy(), dict(key="data.input.img", dtype=np.float32)),
                (OpResizeAndPad2D(), dict(key="data.input.img", resize_to=(2200, 1200), padding=(60, 60))),
                (
                    OpReadDataframe(
                        data_source,
                        key_column="file",
                        key_name="data.input.img_path",
                        columns_to_extract=["ID1", "file", "classification", "subtype"],
                        rename_columns=dict(
                            ID1="data.patientID", classification="data.gt.classification", subtype="data.gt.subtype"
                        ),
                    ),
                    dict(),
                ),
            ],
        )
        return static_pipeline

    @staticmethod
    def dynamic_pipeline(train: bool = False):
        """
        Get suggested dynamic pipeline. including pre-processing that might be modified and augmentation operations.
        :param train : True iff we request dataset for train purpouse
        """
        dynamic_pipeline = PipelineDefault(
            "cmmd_dynamic",
            [
                (OpToTensor(), dict(key="data.input.img", dtype=torch.float32)),
                (OpLambda(partial(torch.unsqueeze, dim=0)), dict(key="data.input.img")),
            ],
        )
        if train:
            dynamic_pipeline.extend(
                [
                    (
                        OpRandApply(OpSample(OpAugAffine2D()), 0.5),
                        dict(
                            key="data.input.img",
                            rotate=Uniform(-30.0, 30.0),
                            scale=Uniform(0.9, 1.1),
                            flip=(RandBool(0.3), RandBool(0.5)),
                            translate=(RandInt(-10, 10), RandInt(-10, 10)),
                        ),
                    ),
                    (
                        OpRandApply(OpSample(OpAugColor()), 0.5),
                        dict(
                            key="data.input.img",
                            gamma=Uniform(0.9, 1.1),
                            contrast=Uniform(0.85, 1.15),
                            mul=Uniform(0.95, 1.05),
                            add=Uniform(-0.06, 0.06),
                        ),
                    ),
                ]
            )
        return dynamic_pipeline

    @staticmethod
    def merge_clinical_data_with_dicom_tags(data_dir: str, data_misc_dir: str, target: str) -> str:
        """
        Creates a csv file that contains label for each image ( instead of patient as in dataset given file)
        by reading metadata ( breast side and view ) from the dicom files and merging it with the input csv
        If the csv already exists , it will skip the creation proccess
        :param data_dir                     dataset root path
        :param data_misc_dir                path to save misc files to be used later
        :return: the new csv file path
        :return: sample ids of used images
        """
        input_source = os.path.join(data_dir, "CMMD_clinicaldata_revision.csv")
        combined_file_path = os.path.join(data_misc_dir, "files_combined.csv")
        if os.path.isfile(combined_file_path):
            print("Found ground truth file:", combined_file_path)
            merged_clinical_data = pd.read_csv(combined_file_path)
            all_sample_ids = merged_clinical_data["file"].to_list()
            return merged_clinical_data, all_sample_ids
        print("Did not find exising ground truth file!")
        Path(data_misc_dir).mkdir(parents=True, exist_ok=True)
        clinical_data = pd.read_csv(input_source)
        scans = []
        for patient in os.listdir(os.path.join(data_dir, "CMMD")):
            path = os.path.join(data_dir, "CMMD", patient)
            for dicom_file in glob.glob(os.path.join(path, "**/*.dcm"), recursive=True):
                file = dicom_file[len(data_dir) + 1 :] if dicom_file.startswith(data_dir) else ""
                dcm = pydicom.dcmread(os.path.join(data_dir, file))
                scans.append(
                    {
                        "ID1": patient,
                        "LeftRight": dcm[0x00200062].value,
                        "file": file,
                        "view": dcm[0x00540220].value.pop(0)[0x00080104].value,
                    }
                )
        dicom_tags = pd.DataFrame(scans)
        merged_clinical_data = pd.merge(clinical_data, dicom_tags, how="outer", on=["ID1", "LeftRight"])
        merged_clinical_data = merged_clinical_data[merged_clinical_data[target].notna()]
        merged_clinical_data["classification"] = np.where(merged_clinical_data["classification"] == "Benign", 0, 1)
        merged_clinical_data["subtype"] = merged_clinical_data["subtype"].replace(
            {"Luminal A": 0, "Luminal B": 1, "HER2-enriched": 2, "triple negative": 3, np.nan: "4"}
        )
        merged_clinical_data = merged_clinical_data.dropna()
        merged_clinical_data.to_csv(combined_file_path)
        all_sample_ids = merged_clinical_data["file"].to_list()
        return merged_clinical_data, all_sample_ids

    @staticmethod
    def dataset(
        data_dir: str,
        data_misc_dir: str,
        target: str,
        cache_dir: str = None,
        reset_cache: bool = True,
        num_workers: int = 10,
        sample_ids: Optional[Sequence[Hashable]] = None,
        train: bool = False,
    ):
        """
        Creates Fuse Dataset single object (either for training, validation and test or user defined set)
        :param data_dir:                    dataset root path
        :param data_misc_dir                path to save misc files to be used later
        :param target                       target name used from the ground truth dataframe
        :param cache_dir:                   Optional, name of the cache folder
        :param reset_cache:                 Optional,specifies if we want to clear the cache first
        :param num_workers: number of processes used for caching
        :param sample_ids: dataset including the specified sample_ids or None for all the samples. sample_id is case_{id:05d} (for example case_00001 or case_00100).
        :param train: True if used for training  - adds augmentation operations to the pipeline
        :return: DatasetDefault object
        """

        input_source_gt, all_sample_ids = CMMD.merge_clinical_data_with_dicom_tags(data_dir, data_misc_dir, target)

        if sample_ids is None:
            sample_ids = all_sample_ids

        static_pipeline = CMMD.static_pipeline(data_dir, input_source_gt, target)
        dynamic_pipeline = CMMD.dynamic_pipeline(train=train)

        cacher = SamplesCacher(
            "cmmd_cache_ver",
            static_pipeline,
            cache_dirs=[cache_dir],
            restart_cache=reset_cache,
            audit_first_sample=False,
            audit_rate=None,
            workers=num_workers,
        )

        my_dataset = DatasetDefault(
            sample_ids=sample_ids,
            static_pipeline=static_pipeline,
            dynamic_pipeline=dynamic_pipeline,
            cacher=cacher,
        )

        my_dataset.create()
        return my_dataset



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""
from abc import ABC, abstractmethod

from typing import Any, Optional, Sequence, List, Tuple
import random

import numpy as np


class ParamSamplerBase(ABC):
    """
    Base class for param sampler
    """

    @abstractmethod
    def sample(self) -> Any:
        """
        :return: the sampled value
        """
        raise NotImplementedError


class Uniform(ParamSamplerBase):
    def __init__(self, min: float, max: float):
        """
        Uniform distribution between min and max
        """
        super().__init__()

        # store input parameters
        self.min = min
        self.max = max

    def sample(self) -> float:
        """
        :return: random float in range [min-max]
        """
        return random.uniform(self.min, self.max)

    def __str__(self):
        return f"Uniform [{self.min} - {self.max}] "


class RandInt(ParamSamplerBase):
    def __init__(self, min: int, max: int):
        """
        Uniform integer distribution between min and max
        """
        super().__init__()

        # store input parameters
        self.min = min
        self.max = max

    def sample(self) -> int:
        """
        :return: random int in range [min-max]
        """
        return random.randint(self.min, self.max)

    def __str__(self):
        return f"RandInt [{self.min} - {self.max}] "


class RandBool(ParamSamplerBase):
    def __init__(self, probability: float):
        """
        Random boolean according to the given probability
        """
        super().__init__()

        # store input parameters
        self.probability = probability

    def sample(self) -> bool:
        """
        :return: random boolean according to probability
        """
        return random.uniform(0, 1) <= self.probability

    def __str__(self):
        return f"RandBool p={self.probability}] "


class Choice(ParamSamplerBase):
    def __init__(self, seq: Sequence, probabilities: Optional[List[float]] = None, k: int = 0):
        """
        Random choice out of a sequence
        Return a k sized list of population elements chosen with replacement
        If k=0 return a single element instead of a list
        """
        super().__init__()

        # store input parameters
        self.seq = seq
        self.probabilities = probabilities
        self.k = k

    def sample(self) -> Any:
        """
        :return: random element according to probabilities
        """
        if self.k == 0:
            return random.choices(self.seq, weights=self.probabilities)[0]
        else:
            return random.choices(self.seq, weights=self.probabilities, k=self.k)

    def __str__(self):
        return f"Choice seq={self.seq}, w={self.probabilities}] "


class Gaussian(ParamSamplerBase):
    """
    Gaussian noise
    """

    def __init__(self, shape: Tuple[int, ...], mean: float, std: float):
        """
        :param shape: patch size of the required noise
        :param mean: mean of the gauss noise
        :param std: std of the gauss noise
        """
        super().__init__()

        # store input parameters
        self.shape = shape
        self.mean = mean
        self.std = std

    def sample(self) -> np.ndarray:
        """
        :return: random gaussian noise
        """
        return self.std * np.random.randn(*list(self.shape)) + self.mean


def draw_samples_recursively(data: Any) -> Any:
    """
    Generate a copy of the data structure, replacing each ParamSamplerBase with a random sample.

    :param data:data_structure: recursively looking for ParamSamplerBase in a dictionary and a sequence
    :return: See above
    """
    # if a dictionary return a copy of the dictionary try to sample each value recursively
    if isinstance(data, dict):
        data_dict: dict = data.copy()
        for key in data_dict:
            data_dict[key] = draw_samples_recursively(data_dict[key])
        return data_dict

    # if a list  return a copy of the list try to sample each element recursively
    if isinstance(data, list):
        data_lst: list = data[:]
        for ii in range(len(data_lst)):
            data_lst[ii] = draw_samples_recursively(data_lst[ii])
        return data_lst

    # if a tuple  return a copy of the tuple try to sample each element recursively
    if isinstance(data, Tuple):
        data_tuple = tuple((draw_samples_recursively(data[ii]) for ii in range(len(data))))
        return data_tuple

    # if ParamSamplerBase, sample a number
    if isinstance(data, ParamSamplerBase):
        data_sampler: ParamSamplerBase = data
        return data_sampler.sample()

    # otherwise return the original data
    return data



from functools import partial
from typing import List, Optional
from fuse.eval.metrics.libs.instance_segmentation import MetricsInstanceSegmentaion

import numpy as np

from fuse.eval.metrics.metrics_common import MetricPerSampleDefault


def average_sample_results(metric_result: List[float], threshold: float = 0.5) -> float:
    """
    Calculates average result per class and average result over classes on a specific metric
    metric_result assumed to have same type of keys as in class_weights which represents the different classes
    :param metric_result: list of per image metric results ,each element is a dictionary where the key is class id and value is the metric score
    :param threshold: a number which determines the metric value of which above we consider for averaging
                      it's purpouse to ignore misdetected instances , 0.5 is the common threshold on iou metric
    :return: average result over images
    """
    per_instane_results = []
    for res in metric_result:
        metric_maximum = res.max(axis=0)
        metric_maximum = metric_maximum[metric_maximum > threshold]
        if len(metric_maximum) > 0:
            per_instane_results.append(metric_maximum.mean())
    result = np.mean(per_instane_results)
    return result


class MetricInstanceIouJaccard(MetricPerSampleDefault):
    """
    Compute aggregated IOU Jaccard score for binary instance segmentation input
    """

    def __init__(
        self,
        pred: str,
        target: str,
        segmentation_pred_type: str,
        segmentation_target_type: str,
        height: str,
        width: str,
        threshold: Optional[float] = 0.5,
        **kwargs
    ):
        """
        See super class for the missing params , to read more about the segmentation types go to class MetricsInstanceSegmentaion
        :param segmentation_pred_type: input pred format - pixel_map , uncompressed_RLE , compressed_RLE , polygon , bbox
        :param segmentation_target_type: input target format - pixel_map , uncompressed_RLE , compressed_RLE , polygon , bbox
        :param height: height of the original image ( y axis)
        :param width: width of the original image ( x axis)
        :param threshold: a number which determines the metric value of which above we consider for averaging
                          it's purpouse to ignore misdetected instances , 0.5 is the common threshold on iou metric
        """
        iou = partial(
            MetricsInstanceSegmentaion.iou_jaccard,
            segmentation_pred_type=segmentation_pred_type,
            segmentation_target_type=segmentation_target_type,
        )
        average = partial(average_sample_results, threshold=threshold)
        super().__init__(
            pred=pred,
            target=target,
            height=height,
            width=width,
            metric_per_sample_func=iou,
            result_aggregate_func=average,
            **kwargs
        )



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import os
from typing import Dict, Hashable, Iterable, List, Optional, OrderedDict, Sequence, Union
import pickle
import pandas as pd

from fuse.utils import read_dataframe
from fuse.utils import NDict
from fuse.utils import CollateToBatchList

from fuse.eval.metrics.metrics_common import MetricBase


class EvaluatorDefault:
    """
    Evaluator
    Combine all the input sources, evaluate using the specified metrics,
    generate a report and returns a dictionary with all the metrics results.
    """

    def __init__(self) -> None:
        pass

    def eval(
        self,
        ids: List[Hashable],
        data: Union[
            str,
            pd.DataFrame,
            Sequence[str],
            Sequence[pd.DataFrame],
            Iterable,
            Dict[str, Union[str, pd.DataFrame, Sequence[str], Sequence[pd.DataFrame]]],
        ],
        metrics: OrderedDict[str, MetricBase],
        id_key: str = "id",
        batch_size: Optional[int] = None,
        output_dir: Optional[str] = None,
    ) -> NDict:
        """
        evaluate, return, print and optionally dump results to a file
        :param ids: list if sample_ids to consider
        :param data: data to evaluate
                     Supported formats:
                        (1) str - path to a file -  csv, pickled dataframe or json. Assumes column called 'id' that holds unique identifier for a sample
                        (2) dataframe - assumes column called 'id'  that holds unique identifier for a sample
                        (3) Sequence of (1) or (2) - each element in the sequence considered to be a different fold.
                                                    ("evaluator_fold" will be added automatically to store the fold number)
                        (4) Dict of (1), (2)  or (3) - each item (key-value) in the dict may contain different information about the same sample.
                                                   A prefix of "<key>." will be automatically added to each column name to avoid from collisions.
                        (5) Iterator - either sample iterator or batch iterator. See param batch_size documentation for more details.

        :param id_key: the column name or key that include the sample id (unique identifier)
        :param batch_size: Optional. Use when the data is too big to read together,
                           Options:
                                None - read all the data together
                                0 - read each batch separately.
                                    data must be an iterator of batch_dict. batch_dict is a dictionary (or nested dictionary).
                                    Each item in this dictionary must include value per sample in the batch.
                                integer > 0 - read each sample separately.
                                              data must be an iterator of samples.
                                              A batch will be automatically created from batch_size samples
        :param output_dir: Optional - dump results to directory
        :return: dictionary that holds all the results.
        """

        if ids is not None:
            ids_df = pd.DataFrame(ids, columns=[id_key]).set_index(id_key)  # use the specified samples
        else:
            ids_df = None  # use all samples

        if batch_size is None:
            data_df = self.read_data(data, ids_df, id_key=id_key)
            data_df["id"] = data_df[id_key]

            # pass data
            for metric_name, metric in metrics.items():
                try:
                    metric.set(data_df)
                except:
                    print(f"Error: metric {metric_name} set() method failed")
                    raise
        elif batch_size > 0:
            # data is iterable - each iteration is a single sample (represented by a dictionary)
            assert isinstance(data, Iterable), "Error: batch mode (batch_size != None) supports only iterable data"
            collate_fn = CollateToBatchList()
            data_iter = iter(data)
            while True:
                samples = []
                try:
                    for _ in range(batch_size):
                        sample = next(data_iter)
                        sample["id"] = sample[id_key]
                        samples.append(sample)
                except StopIteration:
                    break
                finally:
                    if len(samples) > 0:
                        batch = collate_fn(samples)
                        for metric in metrics.values():
                            metric.collect(batch)
        else:
            # data is iterable - each iteration is batch dict (each used element in dict should have batch dimension or list of size batch size)
            assert isinstance(data, Iterable), "Error: batch mode (batch_size != None) supports only iterable data"
            data_iter = iter(data)
            while True:
                try:
                    batch = next(data_iter)
                    for metric in metrics.values():
                        metric.collect(batch)
                except StopIteration:
                    break

        # evaluate
        results = NDict()
        for metric_name, metric in metrics.items():
            try:
                results[f"metrics.{metric_name}"] = metric.eval(results, ids)
            except:
                print(f"Error: metric {metric_name} eval() method failed")
                raise

        # dump results
        self.dump_metrics_results(results, output_dir)

        return results

    def read_data(
        self,
        data: Union[
            str,
            pd.DataFrame,
            Sequence[str],
            Sequence[pd.DataFrame],
            Dict[str, Union[str, pd.DataFrame, Sequence[str], Sequence[pd.DataFrame], Iterable]],
        ],
        ids_df: pd.DataFrame,
        id_key: str,
        error_missing_ids: bool = True,
        error_duplicate: bool = True,
    ):
        """
        Read data and convert to a single dataframe
        :param data: the input data to covert - see eval() method to understand the supported options
        :param ids_df: the returned dataframe will include only the ids specified in ids_df
        :param error_missing_ids: raise an error if data does not include some of the ids specified in ids_df
        :param error_duplicate: raise an error if id exist more than once in data
        """
        if isinstance(data, pd.DataFrame):  # data is already a dataframe
            result_data = data
            # make sure "id" column exist and set it as index
            if id_key not in result_data.keys():
                raise Exception(
                    "Error: 'id' column/key wasn't found in data. Index column specify unique identifier per sample"
                )
            result_data = result_data.set_index(keys=id_key, drop=False)

        elif isinstance(data, str):  # data is path to a file

            result_data = read_dataframe(data)
            # make sure "id" column exist and set it as index
            if id_key not in result_data.keys():
                raise Exception(
                    "Error: 'id' column/key wasn't found in data. Index column specify unique identifier per sample"
                )
            result_data = result_data.set_index(keys=id_key, drop=False)

        elif isinstance(data, Sequence) and isinstance(data[0], (str, pd.DataFrame)):  # data is a sequence of folds
            data_lst = []
            for fold, data_elem in enumerate(data):
                data_elem_df = self.read_data(data_elem, ids_df, error_missing_ids=False, id_key=id_key)
                # add "fold" number to dataframe
                data_elem_df["evaluator_fold"] = fold
                data_lst.append(data_elem_df)
            result_data = pd.concat(data_lst)

        elif isinstance(data, dict):  # data is dictionary of dataframes
            df_list = []
            all_ids = set()
            for key, data_elem in data.items():

                try:
                    data_elem_df = self.read_data(data_elem, ids_df, id_key=id_key)
                    all_ids = all_ids.union(set(data_elem_df.index))
                except:
                    print(f"Error: key={key}")
                    raise

                data_elem_df = data_elem_df.add_prefix(key + ".")
                df_list.append(data_elem_df)

            # make sure ids exists in all dataframes
            for data_elem_df in df_list:
                missing_ids = all_ids - set(data_elem_df.index)
                if len(missing_ids) > 0:
                    raise Exception(
                        f"Error: ids {missing_ids} are missing in data['{data_elem_df.keys()[0].split('.')[0]}']"
                    )

            result_data = pd.concat(df_list, axis=1, join="inner")
            result_data[id_key] = result_data.index

        elif isinstance(data, Iterable):
            result_data = pd.DataFrame([dict(elem) for elem in data])

            # make sure "id" column exist and set it as index
            if id_key not in result_data.keys():
                raise Exception(
                    f"Error: {id_key} column/key wasn't found in data. Index column specify unique identifier per sample - available keys are {result_data.keys()}"
                )
            result_data = result_data.set_index(keys=id_key, drop=False)

        else:
            raise Exception(f"Error: unexpected data type {type(data)}")

        if ids_df is not None:
            # filter to include only the required ids
            result_data = result_data.loc[result_data.index.intersection(ids_df.index)]

            # make sure all ids exists
            if error_missing_ids and len(ids_df) > len(result_data):
                missing_ids = set(ids_df.index) - set(result_data.index)
                raise Exception(f"Error: missing samples in evaluation (missing_ids={missing_ids}")

        # raise error if the same id exist more than once
        if error_duplicate:
            duplicated = result_data.index.duplicated(keep=False)
            duplications = result_data.loc[duplicated].index.unique()
            if not duplications.empty:
                raise Exception(f"Error: found duplicated ids in data: {list(duplications)}")

        return result_data

    def dump_metrics_results(self, metrics_results: NDict, output_dir: str) -> None:
        """
        Dump results to a file
        :param metrics_results: results return from metric.process()
        :param output_dir: directory to dump files to
        :return: None
        """
        # metrics text results
        results = "Results:\n"

        for metric_name in metrics_results["metrics"]:
            # skip metrics use as intermidate computation
            if metric_name.startswith("_"):
                continue
            metric_result = metrics_results["metrics"][metric_name]
            results += f"\nMetric {metric_name}:\n"
            results += "------------------------------------------------\n"

            if isinstance(metric_result, dict):
                metric_result = NDict(metric_result)
                keys = metric_result.keypaths()
                for key in keys:
                    results += f"{key}:\n{metric_result[key]}\n"
            else:
                results += f"{metric_result}\n"

        # print to screen
        print(results)

        # make sure the output folder exist
        if output_dir is not None:
            os.makedirs(output_dir, exist_ok=True)

            # save text results
            with open(os.path.join(output_dir, "results.txt"), "w") as output_file:
                output_file.write(results)

            # save pickled results
            with open(os.path.join(output_dir, "results.pickle"), "wb") as output_file:
                pickle.dump(metrics_results, output_file)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import sys
import logging
import os

from typing import Optional, Union
import pandas as pd
import torch
from torch.utils.data.dataloader import DataLoader
from fuse.data.utils.collates import CollateDefault
from fuseimg.datasets.knight import KNIGHT

from fuse.utils.utils_logger import fuse_logger_start
from fuse.utils.file_io.file_io import save_dataframe

from examples.fuse_examples.imaging.classification.knight.eval.eval import TASK1_CLASS_NAMES, TASK2_CLASS_NAMES
from fuse.dl.lightning.pl_module import LightningModuleDefault
import pytorch_lightning as pl

# add parent directory to path, so that 'knight' folder is treated as a module
sys.path.insert(0, os.path.join(os.path.dirname(os.path.abspath(__file__)), ".."))
from baseline.fuse_baseline import make_model  # noqa


def make_predictions_file(
    model_dir: str,
    model: torch.nn.Module,
    checkpoint: str,
    data_path: str,
    cache_path: Optional[str],
    split: Union[str, dict],
    output_filename: str,
    predictions_key_name: str,
    task_num: int,
):
    """
    Automaitically make prediction files in the requested format - given path to model dir create by FuseMedML during training
    :param model_dir: path to model dir create by FuseMedML during training
    :param model: definition of the model
    :param data_path: path to the original data downloaded from https://github.com/neheller/KNIGHT
    :param cache_path: Optional - path to the cache folder. If none, it will pre-processes the data again
    :param split: either path to pickled dictionary or the actual dictionary specifing the split between train and validation. the dictionary maps "train" to list of sample descriptors and "val" to list of sample descriptions
    :param output_filename: filename of the output csv file
    :param predictions_key_name: the key in batch_dict of the model predictions
    :param task_num: either 1 or 2 (task 1 or task 2)
    """
    # Logger
    fuse_logger_start(console_verbose_level=logging.INFO)
    lgr = logging.getLogger("Fuse")
    lgr.info("KNIGHT: make predictions file in FuseMedML", {"attrs": ["bold", "underline"]})
    lgr.info(f"predictions_filename={os.path.abspath(output_filename)}", {"color": "magenta"})

    # Data
    # read train/val splits file.
    if isinstance(split, str):
        split = pd.read_pickle(split)
        if isinstance(split, list):
            # For this example, we use split 0 out of the the available cross validation splits
            split = split[0]
    if split is None:  # test mode
        json_filepath = os.path.join(data_path, "features.json")
        data = pd.read_json(json_filepath)
        split = {"test": list(data.case_id)}

    dataset = KNIGHT.dataset(data_path, cache_path, split, reset_cache=False)
    dl = DataLoader(
        dataset=dataset,
        shuffle=False,
        drop_last=False,
        batch_sampler=None,
        batch_size=2,
        num_workers=8,
        collate_fn=CollateDefault(),
    )

    pl_module = LightningModuleDefault(
        model_dir=model_dir,
        model=model,
    )

    pl_module.set_predictions_keys([predictions_key_name])

    pl_trainer = pl.Trainer(
        default_root_dir=model_dir,
        accelerator="gpu",
        devices=1,
        strategy=None,
        auto_select_gpus=True,
    )

    predictions = pl_trainer.predict(pl_module, dl, ckpt_path=checkpoint)

    # Convert to required format
    if task_num == 1:
        class_names = TASK1_CLASS_NAMES
    elif task_num == 2:
        class_names = TASK2_CLASS_NAMES
    else:
        raise Exception(f"Unexpected task num {task_num}")

    predictions_score_names = [f"{cls_name}-score" for cls_name in class_names]
    data = []
    for prediction in predictions:
        for i in range(len(prediction["id"])):
            data.append(list(prediction["model.output.head_0"][i]))

    predictions_df = pd.DataFrame(data, columns=list(predictions_score_names))
    predictions_df.reset_index(inplace=True)
    predictions_df.rename({"index": "Case_id"}, axis=1, inplace=True)

    # save file
    save_dataframe(predictions_df, output_filename, index=False)


if __name__ == "__main__":
    """
    Automaitically make prediction files in the requested format - given model definition and path to model dir create by FuseMedML during training
    """
    # no arguments - set arguments inline - see details in function make_predictions_file
    model_dir = ""
    checkpoint = "best"
    data_path = ""
    cache_path = ""
    split = None
    output_filename = "validation_predictions.csv"
    predictions_key_name = "model.output.head_0"
    task_num = 1  # 1 or 2

    use_data = {"imaging": True, "clinical": True}  # specify whether to use imaging, clinical data or both
    num_classes = 2
    imaging_dropout = 0.5
    clinical_dropout = 0.0
    fused_dropout = 0.5

    model = make_model(
        use_data=use_data, num_classes=num_classes, imaging_dropout=imaging_dropout, fused_dropout=fused_dropout
    )

    make_predictions_file(
        model_dir=model_dir,
        model=model,
        checkpoint=checkpoint,
        data_path=data_path,
        cache_path=cache_path,
        split=split,
        output_filename=output_filename,
        predictions_key_name=predictions_key_name,
        task_num=task_num,
    )



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Any

from fuse.utils.misc.misc import Singleton


class FuseDebug(metaclass=Singleton):
    """
    Debug settings. See __init__() for available modes
    """

    def __init__(self, mode: str = "default"):
        """
        :param mode: debug mode that can later be customize by override_setting().
                    Supported modes:
                    - 'default' - no debug tools
                    - 'debug' - best settings to use debugger: disable multiprocessing and multi threading
        """
        # possible values for each attribute
        self._settings_supported_values = {
            # allows to disable multiprocessing and run everything using run_multiprocessed() on the main process instead
            "multiprocessing": ["default", "main_process"],
            # allows to override number of dataloader data workers
            "manager_override_num_dataloader_workers": lambda x: isinstance(x, int) and x >= 0,
            # allows to override number of gpus
            "manager_override_num_gpus": lambda x: isinstance(x, int) and x >= 0,
        }
        # defined the supported modes
        self._modes = {}
        # normal mode
        default_settings = {key: "default" for key in self._settings_supported_values}
        self._modes["default"] = default_settings
        # debug mode
        debug_settings = {
            "multiprocessing": "main_process",
            "manager_override_num_gpus": 1,
            "manager_override_num_dataloader_workers": 0,
        }
        self._modes["debug"] = debug_settings

        self._settings: dict

        self.set_mode(mode)

    def set_mode(self, mode: str) -> None:
        """
        set debug mode
        :param mode: see __init__{} for available modes
        """
        self._settings = self._modes["default"]
        self.override_mode(mode)

    def override_mode(self, mode: str) -> None:
        """
        Override just the settings relevant to this mode
        :param mode: see __init__{} for available modes
        """
        assert mode in self._modes
        if mode != "default":
            for key, value in self._modes[mode].items():
                self.set_setting(key, value)

    def set_setting(self, name: str, value: Any) -> None:
        """
        :param name: setting name. See self._settings_supported_values for available settings.
        :param value: value to override to. See self._settings_supported_values for possible values.
        """
        assert name in self._settings_supported_values, f"setting {name} is not supported"
        supported_values = self._settings_supported_values[name]
        if isinstance(supported_values, list):
            assert (
                value in supported_values
            ), f"value {value} is not supported for setting {name}, supported values are {supported_values}"
        else:
            assert supported_values(value), f"value {value} is not supported for setting {name}"

        self._settings[name] = value

    def get_setting(self, name: str) -> Any:
        """
        :param name: setting name. See self._settings_supported_values for possible settings.
        :return: te value of that setting. See self._settings_supported_values for possible values.
        """
        assert name in self._settings_supported_values, f"setting {name} is not supported"
        return self._settings[name]



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import unittest

import random

from fuse.utils import Seed


class TestSeed(unittest.TestCase):
    """
    Test Seed functionality
    """

    def test_dataloader(self):
        # FIXME: test dataloader after dataset implementation
        pass


if __name__ == "__main__":
    unittest.main()



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""
from typing import Union, List, Optional
from abc import abstractmethod
from fuse.data.utils.sample import get_sample_id
from fuse.utils.ndict import NDict
from fuse.data.ops.hashable_class import HashableClass


class OpBase(HashableClass):
    """
    Operator Base Class
    Operators are the building blocks of the sample processing pipeline.
    Each operator gets as an input the sample_dict as created by the previous operator in pipeline,
    modify sample_dict (can either add/delete/modify fields in sample_dict)before passing it to the next operator in pipeline.
    """

    @abstractmethod
    def __call__(self, sample_dict: NDict, **kwargs) -> Union[None, dict, List[dict]]:
        """
        call function that apply the operation
        :param sample_dict: the generated dictionary generated so far (generated be the previous ops in the pipeline)
                            The first op will typically get just the sample_id stored in sample_dict['data']['sample_id']
        :param kwargs: additional arguments defined per operation
        :return: Typically modified sample_dict.
                There are two special cases supported only if the operation is in static pipeline:
                * return None - ignore the sample and do not raise an error
                * return list of sample_dict - a case splitted to few samples. for example image splitted to patches.
        """
        raise NotImplementedError


class OpReversibleBase(OpBase):
    """
    Special case of op - declaring that the operation can be reversed when required
    (useful to reverse processing steps before presenting the output)
    If there is nothing to reverse - to just declare that the op is reversible inherit from  OpReversibleBase instead of OpBase and implement simple reverse method that returns sample_dict as is.

    If some logic required to reverse the operation:
        (1) record the information required to reverse the operation in __call__ function. Use op_id to store it in sample_dict (sample_dict[op_id] = <information to record?).
        (2) override reverse() method: read the recorded information from sample_dict[op_id] and use it to reverse the operation
    """

    @abstractmethod
    def __call__(self, sample_dict: NDict, op_id: Optional[str], **kwargs) -> Union[None, dict, List[dict]]:
        """
        See OpBase.__call__ for more infomation. The only difference is the extra argument that can be used to "record" the information required to reverse the operation.
        :param op_id: unique identifier for an operation.
                      Might be used to support reverse operation as sample_dict key.
                      In such a case use sample_dict[op_id] = info_to_store

        """
        raise NotImplementedError

    def reverse(self, sample_dict: NDict, key_to_reverse: str, key_to_follow: str, op_id: Optional[str]) -> dict:
        """
        reverse operation
        If a reverse operation is not necessary (for example operator that reads an image),
        implement simple reverse method that returns sample_dict as is

        If reverse operation is necessary but not required by the project so far,
        inherit from OpBase (will throw an NotImplementedError in case the reverse operation will be called).

        To support reverse operation, store the parameters which necessary to apply the reverse operation
        such as key to the transformed value and the argument to the transform operation in sample_dict[op_id].
        Those values can be extracted back during the reverse operation.

        :param sample_dict: the dictionary as modified by the previous steps (reversed direction)
        :param op_id: See op_id in __call__ function
        :param key_to_reverse: the required value to reverse
        :param key_to_follow: run the reverse according to the operation applied on this value
        :return: modified sample_dict
        """
        raise NotImplementedError(
            f"op {self} is not reversible. If there is nothing to reverse, just implement simple reverse method that returns sample_dict as is. If extra logic required to reverse follow the instructions in OpReversibleBase"
        )


def op_call(op: OpBase, sample_dict: NDict, op_id: str, **kwargs):
    try:
        if isinstance(op, OpReversibleBase):
            return op(sample_dict, op_id=op_id, **kwargs)
        else:  # OpBase but note reversible
            return op(sample_dict, **kwargs)
    except:
        # error messages are cryptic without this. For example, you can get "TypeError: __call__() got an unexpected keyword argument 'key_out_input'" , without any reference to the relevant op!
        print(
            "************************************************************************************************************************************\n"
            + f"error in __call__ method of op={op}, op_id={op_id}, sample_id={get_sample_id(sample_dict)} - more details below"
            + "*************************************************************************************************************************************\n"
        )
        raise


def op_reverse(op, sample_dict: NDict, key_to_reverse: str, key_to_follow: str, op_id: Optional[str]):
    if isinstance(op, OpReversibleBase):
        try:
            return op.reverse(sample_dict, key_to_reverse, key_to_follow, op_id)
        except:
            # error messages are cryptic without this. For example, you can get "TypeError: __call__() got an unexpected keyword argument 'key_out_input'" , without any reference to the relevant op!
            print(
                f"error in reverse method of op={op}, op_id={op_id}, sample_id={get_sample_id(sample_dict)} - more details below"
            )
            raise

    else:  # OpBase but note reversible
        raise NotImplementedError(
            f"op {op} is not reversible. If there is nothing to reverse, just inherit OpReversibleBase instead of OpBase and implement simple reverse method that returns sample_dict as is. If extra logic required to reverse follow the instructions in OpReversibleBase"
        )



from functools import partial
from typing import Dict, List, Optional
from collections import defaultdict
from fuse.eval.metrics.libs.segmentation import MetricsSegmentation

import numpy as np

from fuse.eval.metrics.metrics_common import MetricPerSampleDefault


def average_sample_results(
    metric_result: List[Dict[int, float]], class_weights: Optional[Dict[int, float]] = None
) -> Dict[str, float]:
    """
    Calculates average result per class and average result over classes on a specific metric
    metric_result assumed to have same type of keys as in class_weights which represents the different classes
    :param metric_result: list of per image metric results ,each element is a dictionary where the key is class id and value is the metric score
    :param class_weights: weight per segmentation class , we assume sum of total weights is 1 and each element is in 0-1 range
    :return: dictionary of average result per class and average result over classes
    """
    average_results = {}
    aggregated_results = defaultdict(list)
    for sample in metric_result:
        for key, score in sample.items():
            aggregated_results[key].append(score)
    total_avarage = 0
    if class_weights is None:
        for key, result_list in aggregated_results.items():
            average_results[key] = np.sum(result_list) / len(result_list)
            total_avarage += average_results[key]
    else:
        for key, result_list in aggregated_results.items():
            average_results[key] = np.sum(result_list) / len(result_list)
            total_avarage += class_weights[key] * average_results[key]
    average_results["average"] = total_avarage / len(aggregated_results)
    return average_results


class MetricDice(MetricPerSampleDefault):
    """
    Compute similarity dice score (2*|X&Y| / (|X|+|Y|)) for every label
    """

    def __init__(
        self,
        pred: str,
        target: str,
        pixel_weight: Optional[str] = None,
        class_weights: Optional[Dict[int, float]] = None,
        **kwargs
    ):
        """
        See super class for the missing params
        used for sematric and binary segmentation
        :param class_weights: weight per segmentation class , we assume sum of total weights is 1 and each element is in 0-1 range
        :param pixel_weight: Optional dictionary key to collect
        """
        average = partial(average_sample_results, class_weights=class_weights)
        super().__init__(
            pred=pred,
            target=target,
            pixel_weight=pixel_weight,
            metric_per_sample_func=MetricsSegmentation.dice,
            result_aggregate_func=average,
            **kwargs
        )


class MetricIouJaccard(MetricPerSampleDefault):
    """
    Compute IOU Jaccard score for every label
    used for sematric and binary segmentation
    """

    def __init__(
        self,
        pred: str,
        target: str,
        pixel_weight: Optional[str] = None,
        class_weights: Optional[Dict[int, float]] = None,
        **kwargs
    ):
        """
        See super class for the missing params
        :param class_weights: weight per segmentation class , we assume sum of total weights is 1 and each element is in 0-1 range
        :param pixel_weight: Optional dictionary key to collect
        """
        average = partial(average_sample_results, class_weights=class_weights)
        super().__init__(
            pred,
            target,
            pixel_weight=pixel_weight,
            metric_per_sample_func=MetricsSegmentation.iou_jaccard,
            result_aggregate_func=average,
            **kwargs
        )


class MetricOverlap(MetricPerSampleDefault):
    """
    Compute overlap score for every label
    used for sematric and binary segmentation
    """

    def __init__(
        self,
        pred: str,
        target: str,
        pixel_weight: Optional[str] = None,
        class_weights: Optional[Dict[int, float]] = None,
        **kwargs
    ):
        """
        See super class for the missing params
        :param class_weights: weight per segmentation class , we assume sum of total weights is 1 and each element is in 0-1 range
        :param pixel_weight: Optional dictionary key to collect
        """
        average = partial(average_sample_results, class_weights=class_weights)
        super().__init__(
            pred,
            target,
            pixel_weight=pixel_weight,
            metric_per_sample_func=MetricsSegmentation.overlap,
            result_aggregate_func=average,
            **kwargs
        )


class Metric2DHausdorff(MetricPerSampleDefault):
    """
    Compute Hausdorff score for every label - works for 2D array only!
    used for sematric and binary segmentation
    """

    def __init__(self, pred: str, target: str, class_weights: Optional[Dict[int, float]] = None, **kwargs):
        """
        See super class for the missing params
        :param class_weights: weight per segmentation class , we assume sum of total weights is 1 and each element is in 0-1 range
        """
        average = partial(average_sample_results, class_weights=class_weights)
        super().__init__(
            pred,
            target,
            metric_per_sample_func=MetricsSegmentation.hausdorff_2d_distance,
            result_aggregate_func=average,
            **kwargs
        )


class MetricPixelAccuracy(MetricPerSampleDefault):
    """
    Compute pixel accuracy score for every label
    used for sematric and binary segmentation
    """

    def __init__(
        self,
        pred: str,
        target: str,
        pixel_weight: Optional[str] = None,
        class_weights: Optional[Dict[int, float]] = None,
        **kwargs
    ):
        """
        See super class for the missing params
        :param class_weights: weight per segmentation class , we assume sum of total weights is 1 and each element is in 0-1 range
        :param pixel_weight: Optional dictionary key to collect
        """
        average = partial(average_sample_results, class_weights=class_weights)
        super().__init__(
            pred,
            target,
            pixel_weight=pixel_weight,
            metric_per_sample_func=MetricsSegmentation.pixel_accuracy,
            result_aggregate_func=average,
            **kwargs
        )



import os
from zipfile import ZipFile
from fuse.utils.file_io.file_io import create_dir
import wget
from typing import Hashable, Optional, Sequence, List, Tuple
import torch

from fuse.data import DatasetDefault
from fuse.data.ops.ops_cast import OpToTensor
from fuse.data.utils.sample import get_sample_id
from fuse.data.pipelines.pipeline_default import PipelineDefault
from fuse.data.ops.op_base import OpBase
from fuse.data.datasets.caching.samples_cacher import SamplesCacher
from fuse.data.ops.ops_aug_common import OpSample
from fuse.data.ops.ops_read import OpReadDataframe
from fuse.data.ops.ops_common import OpLambda, OpOverrideNaN
from fuseimg.data.ops.color import OpToRange

from fuse.utils import NDict

from fuseimg.data.ops.image_loader import OpLoadImage
from fuseimg.data.ops.aug.color import OpAugColor, OpAugGaussian
from fuseimg.data.ops.aug.geometry import OpResizeTo, OpAugAffine2D
from fuse.utils.rand.param_sampler import Uniform, RandInt, RandBool


class OpISICSampleIDDecode(OpBase):
    def __call__(self, sample_dict: NDict) -> NDict:
        """
        decodes sample id into image file name
        """
        sid = get_sample_id(sample_dict)

        img_filename_key = "data.input.img_path"
        sample_dict[img_filename_key] = sid + ".jpg"

        return sample_dict


def derive_label(sample_dict: NDict) -> NDict:
    """
    Takes the sample's ndict with the labels as key:value and assigns to sample_dict['data.label'] the index of the sample's class.
    Also delete all the labels' keys from sample_dict.

    for example:
        If the sample contains {'MEL': 0, 'NV': 1, 'BCC': 0, 'AK': 0, ... }
        will assign, sample_dict['data.label'] = 1 ('NV's index).
        Afterwards the sample_dict won't contain the class' names & values.
    """
    classes_names = ["MEL", "NV", "BCC", "AK", "BKL", "DF", "VASC", "SCC"]

    label = 0
    for idx, cls_name in enumerate(classes_names):
        if int(sample_dict[f"data.cls_labels.{cls_name}"]) == 1:
            label = idx

    sample_dict["data.label"] = label
    return sample_dict


class ISIC:
    """
    ISIC 2019 challenge to classify dermoscopic images and clinical data among nine different diagnostic categories.
    """

    # bump whenever the static pipeline modified
    DATASET_VER = 0

    CLASS_NAMES = ["MEL", "NV", "BCC", "AK", "BKL", "DF", "VASC", "SCC"]

    @staticmethod
    def download(data_path: str, sample_ids_to_download: Optional[Sequence[str]] = None) -> None:
        """
        Download images and metadata from ISIC challenge.
        Doesn't download again if data exists.
        :param data_path: location to store the data
        :param sample_ids_to_downlad: use if you want to download subset of the data. None will download and extract all the data.
        """
        create_dir(data_path)
        create_dir(os.path.join(data_path, "ISIC2019"))
        path = os.path.join(data_path, "ISIC2019/ISIC_2019_Training_Input")
        if not os.path.exists(path):
            print("\nExtract ISIC-2019 training input ... (this may take a few minutes)")

            url = "https://isic-challenge-data.s3.amazonaws.com/2019/ISIC_2019_Training_Input.zip"
            filename = os.path.join(data_path, "ISIC_2019_Training_Input.zip")
            print(f"Extract to {path}")
            if not os.path.exists(filename):
                wget.download(url, os.path.join(data_path, "ISIC_2019_Training_Input.zip"))

            if sample_ids_to_download is not None:
                members = [os.path.join("ISIC_2019_Training_Input", m + ".jpg") for m in sample_ids_to_download]
            else:
                members = None

            with ZipFile(filename, "r") as zipObj:
                # Extract all the contents of zip file in current directory
                zipObj.extractall(path=os.path.join(data_path, "ISIC2019"), members=members)

            print("Extracting ISIC-2019 training input: done")

        path = os.path.join(data_path, "ISIC2019/ISIC_2019_Training_GroundTruth.csv")

        if not os.path.exists(path):
            print("\nExtract ISIC-2019 training gt ... (this may take a few minutes)")

            url = "https://isic-challenge-data.s3.amazonaws.com/2019/ISIC_2019_Training_GroundTruth.csv"
            wget.download(url, os.path.join(data_path, "ISIC2019/ISIC_2019_Training_GroundTruth.csv"))

            print("Extracting ISIC-2019 training gt: done")

        path = os.path.join(data_path, "ISIC2019/ISIC_2019_Training_Metadata.csv")

        if not os.path.exists(path):
            print("\nExtract ISIC-2019 metadata ... (this may take a few minutes)")

            url = "https://isic-challenge-data.s3.amazonaws.com/2019/ISIC_2019_Training_Metadata.csv"
            wget.download(url, os.path.join(data_path, "ISIC2019/ISIC_2019_Training_Metadata.csv"))

            print("Extracting ISIC-2019 metadata: done")

    @staticmethod
    def sample_ids(data_path: str) -> List[str]:
        """
        Gets the samples ids in trainset.
        """
        images_path = os.path.join(data_path, "ISIC2019/ISIC_2019_Training_Input")

        samples = [f.split(".")[0] for f in os.listdir(images_path) if f.split(".")[-1] == "jpg"]
        return samples

    @staticmethod
    def static_pipeline(data_path: str) -> PipelineDefault:
        rename_cls_labels = {c: f"data.cls_labels.{c}" for c in ISIC.CLASS_NAMES}
        rename_cls_labels["image"] = "data.cls_labels.sample_id"  # also extract image (sample_id)
        rename_metadata = {
            "age_approx": "data.input.clinical.age_approx",
            "anatom_site_general": "data.input.clinical.anatom_site_general",
            "sex": "data.input.clinical.sex",
            "image": "data.input.clinical.sample_id",
        }

        static_pipeline = PipelineDefault(
            "static",
            [
                # Decoding sample ID
                (OpISICSampleIDDecode(), dict()),
                # Load Image
                (
                    OpLoadImage(os.path.join(data_path, "ISIC2019/ISIC_2019_Training_Input")),
                    dict(key_in="data.input.img_path", key_out="data.input.img"),
                ),
                # Normalize Images to range [0, 1]
                (OpToRange(), dict(key="data.input.img", from_range=(0, 255), to_range=(0, 1))),
                # Read labels into sample_dict. Each class will have a different entry.
                (
                    OpReadDataframe(
                        data_filename=os.path.join(data_path, "ISIC2019/ISIC_2019_Training_GroundTruth.csv"),
                        key_column="data.cls_labels.sample_id",
                        columns_to_extract=list(rename_cls_labels.keys()),
                        rename_columns=rename_cls_labels,
                    ),
                    dict(),
                ),
                # Read metadata into sample_dict
                (
                    OpReadDataframe(
                        data_filename=os.path.join(data_path, "ISIC2019/ISIC_2019_Training_Metadata.csv"),
                        key_column="data.input.clinical.sample_id",
                        columns_to_extract=list(rename_metadata.keys()),
                        rename_columns=rename_metadata,
                    ),
                    dict(),
                ),
                (OpOverrideNaN(), dict(key="data.input.clinical.anatom_site_general", value_to_fill="N/A")),
                (OpOverrideNaN(), dict(key="data.input.clinical.sex", value_to_fill="N/A")),
                (OpOverrideNaN(), dict(key="data.input.clinical.age_approx", value_to_fill=-1.0)),
                # Squeeze labels into sample_dict['data.label']
                (OpLambda(func=derive_label), dict()),
            ],
        )
        return static_pipeline

    @staticmethod
    def dynamic_pipeline(
        train: bool = False, append: Optional[Sequence[Tuple[OpBase, dict]]] = None
    ) -> PipelineDefault:
        """
        Get suggested dynamic pipeline. including pre-processing that might be modified and augmentation operations.
        :param train: add augmentation if True
        :param append: pipeline steps to append at the end of the suggested pipeline
        """

        dynamic_pipeline = [
            # Resize images to 300x300x3
            (
                OpResizeTo(channels_first=True),
                dict(key="data.input.img", output_shape=(300, 300, 3), mode="reflect", anti_aliasing=True),
            ),
            # Convert to tensor for the augmentation process
            (OpToTensor(), dict(key="data.input.img", dtype=torch.float)),
        ]

        if train:
            dynamic_pipeline += [
                # Augmentation
                (
                    OpSample(OpAugAffine2D()),
                    dict(
                        key="data.input.img",
                        rotate=Uniform(-180.0, 180.0),
                        scale=Uniform(0.9, 1.1),
                        flip=(RandBool(0.3), RandBool(0.3)),
                        translate=(RandInt(-50, 50), RandInt(-50, 50)),
                    ),
                ),
                # Color augmentation
                (
                    OpSample(OpAugColor()),
                    dict(
                        key="data.input.img",
                        gamma=Uniform(0.9, 1.1),
                        contrast=Uniform(0.85, 1.15),
                        add=Uniform(-0.06, 0.06),
                        mul=Uniform(0.95, 1.05),
                    ),
                ),
                # Add Gaussian noise
                (OpAugGaussian(), dict(key="data.input.img", std=0.03)),
            ]

        if append is not None:
            dynamic_pipeline += append

        return PipelineDefault("dynamic", dynamic_pipeline)

    @staticmethod
    def dataset(
        data_path: str,
        cache_path: str,
        train: bool = False,
        reset_cache: bool = False,
        num_workers: int = 10,
        append_dyn_pipeline: Optional[Sequence[Tuple[OpBase, dict]]] = None,
        samples_ids: Optional[Sequence[Hashable]] = None,
    ) -> DatasetDefault:
        """
        Get cached dataset
        :param train: if true returns the train dataset, else the validation one.
        :param reset_cache: set to True to reset the cache
        :param num_workers: number of processes used for caching
        :param append_dyn_pipeline: pipeline steps to append at the end of the suggested dynamic pipeline
        :param sample_ids: dataset including the specified sample_ids or None for all the samples.
        """
        # Download data if doesn't exist
        ISIC.download(data_path=data_path, sample_ids_to_download=samples_ids)

        if samples_ids is None:
            samples_ids = ISIC.sample_ids(data_path)

        static_pipeline = ISIC.static_pipeline(data_path)
        dynamic_pipeline = ISIC.dynamic_pipeline(train, append=append_dyn_pipeline)

        cacher = SamplesCacher(
            f"isic_cache_ver{ISIC.DATASET_VER}",
            static_pipeline,
            [cache_path],
            restart_cache=reset_cache,
            workers=num_workers,
        )

        my_dataset = DatasetDefault(
            sample_ids=samples_ids, static_pipeline=static_pipeline, dynamic_pipeline=dynamic_pipeline, cacher=cacher
        )

        my_dataset.create()
        return my_dataset



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import unittest


from fuse.utils.ndict import NDict
import numpy
import torch
import copy


class TestNDict(unittest.TestCase):
    """
    Test NDict class
    """

    def setUp(self):
        self.nested_dict = NDict({"a": 1, "b": {"c": 2, "d": 3}, "c": 4})

    def test_get(self):
        self.assertEqual(self.nested_dict["a"], 1)
        self.assertEqual(self.nested_dict["b.c"], 2)
        self.assertEqual(self.nested_dict["c"], 4)
        self.assertDictEqual(self.nested_dict["b"], {"c": 2, "d": 3})

    def test_set(self):
        self.nested_dict["a"] = 7
        self.assertEqual(self.nested_dict["a"], 7)
        self.nested_dict["a"] = 1
        self.assertEqual(self.nested_dict["a"], 1)

        self.nested_dict["b.d"] = 5
        self.assertEqual(self.nested_dict["b"]["d"], 5)
        self.nested_dict["b.d"] = 3
        self.assertEqual(self.nested_dict["b"]["d"], 3)

        self.nested_dict["b.e.f.g"] = 9
        self.assertEqual(self.nested_dict["b"]["e"]["f"]["g"], 9)
        del self.nested_dict["b"]["e"]

    def test_pop(self):
        self.nested_dict.pop("b.d")
        all_keys = self.nested_dict.keypaths()
        self.assertSetEqual(set(all_keys), {"a", "b.c", "c"})

        self.nested_dict.pop("c")
        all_keys = self.nested_dict.keypaths()
        self.assertSetEqual(set(all_keys), {"a", "b.c"})
        with self.assertRaises(KeyError):
            self.nested_dict.pop("c")
        with self.assertRaises(KeyError):
            self.nested_dict.pop("lala")

    def test_keypaths(self):
        all_keys = self.nested_dict.keypaths()
        self.assertSetEqual(set(all_keys), {"a", "b.c", "b.d", "c"})

    def test_is_in(self):
        self.assertTrue("a" in self.nested_dict)
        self.assertTrue("b.c" in self.nested_dict)
        self.assertTrue("b.d" in self.nested_dict)
        self.assertTrue("c" in self.nested_dict)
        self.assertFalse("d" in self.nested_dict)
        self.assertFalse("e" in self.nested_dict)

    def test_apply_on_all(self):
        nested_dict_copy = self.nested_dict.clone()

        def plus_one(val: int) -> int:
            return val + 1

        nested_dict_copy.apply_on_all(plus_one)

        # verify values
        all_keys = sorted(self.nested_dict.keypaths())
        all_keys_2 = sorted(nested_dict_copy.keypaths())
        self.assertListEqual(all_keys, all_keys_2)
        for key in all_keys:
            self.assertEqual(self.nested_dict[key] + 1, nested_dict_copy[key])

    def test_flatten(self):
        flat_dict = self.nested_dict.flatten()

        # verify
        all_keys = sorted(self.nested_dict.keypaths())
        flat_keys = sorted(flat_dict.keys())
        self.assertListEqual(all_keys, flat_keys)
        for key in flat_keys:
            self.assertEqual(self.nested_dict[key], flat_dict[key])

    def test_indices(self):
        nested_dict = NDict(
            {
                "a": numpy.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]]),
                "b": {"c": torch.tensor([[10, 11, 12], [13, 14, 15], [16, 17, 18]]), "d": [0, 1, 2], "f": 4},
            }
        )
        indices = numpy.array([True, False, True])
        nested_dict_indices = nested_dict.indices(indices)

        # verify
        self.assertTrue((nested_dict_indices["a"] == numpy.array([[0, 1, 2], [6, 7, 8]])).all())
        self.assertTrue((nested_dict_indices["b.c"] == torch.tensor([[10, 11, 12], [16, 17, 18]])).all())
        self.assertTrue((nested_dict_indices["b.d"] == [0, 2]))
        self.assertTrue((nested_dict_indices["b.f"] == 4))

    def tearDown(self):
        delattr(self, "nested_dict")


if __name__ == "__main__":
    unittest.main()



import os


def change_extension(filepath: str, new_extension: str) -> str:
    """
    modifies [filepath] extension to be [new_extension]

    for example _change_extension('/a/b/c.sdf/ew/some_file.zip', '7zip')
    will return '/a/b/c.sdf/ew/some_file.7zip'
    """

    _basename = os.path.basename(filepath)
    last_dot = _basename.rfind(".")
    if last_dot < 0:
        return filepath + "." + new_extension
    ans = os.path.join(os.path.dirname(filepath), _basename[:last_dot] + "." + new_extension)
    return ans


def get_extension(filepath: str) -> str:
    """
    Returns the extension.
    For example - get_extension('/a/b/c/d/banana.txt') will return '.txt'
    """
    _basename = os.path.basename(filepath)
    last_dot = _basename.rfind(".")
    if last_dot < 0:
        return ""
    ans = _basename[last_dot:]
    return ans


def remove_extension(filepath: str) -> str:
    """
    Returns the filename without the extension
    For example: remove_extension('/a/b/c/d/asdf.txt') with return '/a/b/c/d/asdf'
    """

    _dirname = os.path.dirname(filepath)
    _basename = os.path.basename(filepath)
    if "." not in _basename:
        return filepath

    last_dot = _basename.rfind(".")
    ans = os.path.join(_dirname, _basename[:last_dot])
    return ans



from abc import abstractmethod
from typing import Hashable, List, Sequence, Optional
from fuse.data.utils.sample import get_sample_id
from fuse.utils import NDict
from fuse.data import OpBase
import numpy
import torch


class OpDebugBase(OpBase):
    """
    Base class for debug operations.
    Provides the ability to limit samples to debug (will debug the first k samples).
    Inherits and implements self.call_debug instead of self.__call__.
    """

    def __init__(self, name: Optional[str] = None, sample_ids: Optional[List[Hashable]] = None, num_samples: int = 0):
        """
        :param name: string identifier - might be useful when the debug op display or save information into a file
        :param sample_ids: apply for the specified sample ids. To apply for all set to None.
        :param num_samples: apply for the first num_samples (per process). if 0, will apply for all.
        """
        super().__init__()
        self._name = name
        self._sample_ids = sample_ids
        self._num_samples = num_samples
        self._num_samples_done = 0

    def reset(self, name: Optional[str] = None):
        """Reset operation state"""
        self._num_samples_done = 0
        self._name = name

    def should_debug_sample(self, sample_dict: NDict) -> bool:
        if self._num_samples and self._num_samples_done >= self._num_samples:
            return False

        if self._sample_ids is not None:
            sid = get_sample_id(sample_dict)
            if sid not in self._sample_ids:
                return False

        self._num_samples_done += 1
        return True

    def __call__(self, sample_dict: NDict, **kwargs) -> NDict:
        if self.should_debug_sample(sample_dict):
            self.call_debug(sample_dict, **kwargs)
        return sample_dict

    @abstractmethod
    def call_debug(self, sample_dict: NDict, **kwargs) -> None:
        """The actual debug op implementation"""
        raise NotImplementedError


class OpPrintKeys(OpDebugBase):
    """
    Print list of available keys at a given point in the data pipeline
    It's recommended, but not a must, to run it in a single process.
    ```
    from fuse.utils.utils_debug import FuseDebug
    FuseDebug("debug")
    ```

    Example:
    ```
    (OpPrintKeys(num_samples=1), dict()),
    ```
    """

    def call_debug(self, sample_dict: NDict) -> None:
        print(f"Sample {get_sample_id(sample_dict)} keys:")
        for key in sample_dict.keypaths():
            print(f"\t{key}")


class OpPrintShapes(OpDebugBase):
    """
    Print the shapes/length of every torch tensor / numpy array / sequence
    Add at the top your script to force single process:
    ```
    from fuse.utils.utils_debug import FuseDebug
    FuseDebug("debug")
    ```
    Example:
    ```
    (OpPrintShapes(num_samples=1), dict()),
    ```
    """

    def call_debug(self, sample_dict: NDict) -> None:
        print(f"Sample {get_sample_id(sample_dict)} shapes:")
        for key in sample_dict.keypaths():
            value = sample_dict[key]
            if isinstance(value, torch.Tensor):
                print(f"\t{key} is tensor with shape: {value.shape}")
            elif isinstance(value, numpy.ndarray):
                print(f"\t{key} is numpy array with shape: {value.shape}")
            elif not isinstance(value, str) and isinstance(value, Sequence):
                print(f"\t{key} is sequence with length: {len(value)}")


class OpPrintTypes(OpDebugBase):
    """
    Print the the type of each key

    Add at the top your script to force single process:
    ```
    from fuse.utils.utils_debug import FuseDebug
    FuseDebug("debug")
    ```
    Example:
    ```
    (OpPrintTypes(num_samples=1), dict()),
    ```
    """

    def call_debug(self, sample_dict: NDict) -> None:
        print(f"Sample {get_sample_id(sample_dict)} types:")
        for key in sample_dict.keypaths():
            value = sample_dict[key]
            print(f"\t{key} - {type(value).__name__}")


class OpPrintKeysContent(OpDebugBase):
    """
    Print sample's keys and contents.
    It's recommended, but not a must, to run it in a single process.
    ```
    from fuse.utils.utils_debug import FuseDebug
    FuseDebug("debug")
    ```
    Example:
    ```
    (OpPrintKeysContent(num_samples=1), dict(keys=["data.input.mri_path", "data.label"])),
    (OpPrintKeysContent(num_samples=1), dict(keys=None)),
    ```
    """

    def call_debug(self, sample_dict: NDict, keys: List[str]) -> None:
        """
        :param keys: List of keys to print. Set to 'None' to print all keys.
        """
        print(f"OpPrintKeysContent, sample '{get_sample_id(sample_dict)}', <key> = <content>:")
        dict_keys = sample_dict.keypaths()

        if keys is None:
            keys = dict_keys  # print all keys

        for key in keys:
            if key in dict_keys:
                print(f"\t{key} = {sample_dict[key]}")
            else:
                print(f"\tWARNING! {key} not in sample_dict")



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""
import pathlib
import shutil
import tempfile
import unittest
import os
from fuse.utils.file_io.file_io import create_dir
from fuse.utils.multiprocessing.run_multiprocessed import run_in_subprocess
import sys
import wget

# add parent directory to path, so that 'knight' folder is treated as a module
sys.path.insert(0, os.path.join(os.path.dirname(os.path.abspath(__file__)), ".."))
from imaging.classification.knight.eval.eval import eval
from imaging.classification.knight.make_targets_file import make_targets_file
import imaging.classification.knight.baseline.fuse_baseline as baseline


class KnightTestTestCase(unittest.TestCase):
    def setUp(self):
        self.root = tempfile.mkdtemp()

    def test_eval(self):
        dir_path = pathlib.Path(__file__).parent.resolve()
        target_filename = os.path.join(dir_path, "../imaging/classification/knight/eval/example/example_targets.csv")
        task1_prediction_filename = os.path.join(
            dir_path, "../imaging/classification/knight/eval/example/example_task1_predictions.csv"
        )
        task2_prediction_filename = os.path.join(
            dir_path, "../imaging/classification/knight/eval/example/example_task2_predictions.csv"
        )
        eval(
            target_filename=target_filename,
            task1_prediction_filename=task1_prediction_filename,
            task2_prediction_filename=task2_prediction_filename,
            output_dir=self.root,
        )

    def test_make_targets(self):
        dir_path = pathlib.Path(__file__).parent.resolve()
        data_path = os.path.join(self.root, "data")
        create_dir(data_path)
        cache_path = os.path.join(self.root, "cache")
        output_filename = os.path.join(self.root, "output/validation_targets.csv")
        wget.download(
            "https://raw.github.com/neheller/KNIGHT/main/knight/data/knight.json",
            data_path,
        )
        split = os.path.join(dir_path, "../imaging/classification/knight/baseline/splits_final.pkl")
        create_dir(os.path.dirname(output_filename))
        make_targets_file(data_path=data_path, split=split, output_filename=output_filename)

    @unittest.skipIf("KNIGHT_DATA" not in os.environ, "define environment variable 'KNIGHT_DATA' to run this test")
    def test_train(self):
        os.environ["KNIGHT_CACHE"] = os.path.join(self.root, "train", "cache")
        os.environ["KNIGHT_RESULTS"] = os.path.join(self.root, "train", "results")
        config = """
        experiment_num : 0
        task_num : task_1 # task_1 or task_2 
        num_gpus : 1
        use_data : {"imaging": True, "clinical": True} 
        batch_size : 2
        resize_to : [70, 256, 256]
        num_epochs : 0
        learning_rate : 0.0001
        imaging_dropout : 0.5
        fused_dropout : 0.5
        testing : True

        task_1:
            num_classes : 2
            target_name : "data.gt.gt_global.task_1_label"
            target_metric : "validation.metrics.auc"
        """
        cfg_path = os.path.join(self.root, "test_cfg.yaml")
        with open(cfg_path, "w") as file:
            file.write(config)

        run_in_subprocess(baseline.main, cfg_path, timeout=4000)

    def tearDown(self):
        # Delete temporary directories
        shutil.rmtree(self.root)


if __name__ == "__main__":
    unittest.main()



import cProfile
import pstats
from io import StringIO


class Profiler:
    """
    A simple to use cpu profiler, helps to find CPU bottlenecks

    all you need to do is wrap the code you want to analyze:

    with Profiler('blahblah'):
        #... some code

    Usage example:

    #we create two functions, one is quick and one is very slow

    from itertools import combinations
    def dummy_slow():
        N = 140 #1000
        K = 2
        for _ in range(100):
            all_combs = combinations(range(N),K)
            all_combs = list(all_combs)
        return all_combs

    def dummy_quick():
        a = 1000
        a+=1
        return a

    def get_combs():
        for _ in range(10):
            x = dummy_slow()
            y = dummy_quick()

        return x,y

    with Profiler('banana'):
        get_combs()

    #results in

    starting profiling for banana...
    profiling results for banana :
            23 function calls in 0.779 seconds

    Ordered by: cumulative time

    ncalls  tottime  percall  cumtime  percall filename:lineno(function)
    1    0.001    0.001    0.779    0.779 /some/path/test_profiler.py:22(get_combs)
    10   0.778    0.078    0.778    0.078 /some/path/test_profiler.py:9(dummy_slow)
    1    0.000    0.000    0.000    0.000 /some/path/profiler.py:45(__exit__)
    10   0.000    0.000    0.000    0.000 /some/path/test_profiler.py:17(dummy_quick)
    1    0.000    0.000    0.000    0.000 {method 'disable' of '_lsprof.Profiler' objects}

    #analyzing it we can see that dummy_slow is the bottelneck within get_combs,
    as its cumulative time is 0.778 secs out of the total of 0.779 seconds of get_combs
    on the other hand, dummy_quick is clearly not the bottleneck.

    """

    def __init__(self, txt: str = ""):
        self.txt = txt

    def __enter__(self):
        print(f"starting profiling for {self.txt}...")
        self.pr = cProfile.Profile()
        self.pr.enable()

    def __exit__(self, *args):
        self.pr.disable()
        print(f"profiling results for {self.txt} :")
        self.s = StringIO()
        sortby = "cumulative"
        self.ps = pstats.Stats(self.pr, stream=self.s).sort_stats(sortby).reverse_order()
        self.ps.print_stats()
        print(self.s.getvalue())



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""
from typing import Any, Callable, Dict, List, Sequence

import numpy as np
import torch
from torch.utils.data._utils.collate import default_collate
import torch.nn.functional as F

from fuse.utils import NDict
from fuse.utils.data.collate import CollateToBatchList
from fuse.data import get_sample_id_key


class CollateDefault(CollateToBatchList):
    """
    Default collate_fn method to be used when creating a DataLoader.
    Will collate each value with PyTorch default collate.
    Special collates per key can be specified in special_handlers_keys
    sample_id key will be collected to a list.
    Few options to special handlers implemented in this class as static methods
    """

    def __init__(
        self,
        skip_keys: Sequence[str] = tuple(),
        raise_error_key_missing: bool = True,
        special_handlers_keys: Dict[str, Callable] = None,
    ):
        """
        :param skip_keys: do not collect the listed keys
        :param special_handlers_keys: per key specify a callable which gets as an input list of values and convert it to a batch.
                                      The rest of the keys will be converted to batch using PyTorch default collate_fn()
        :param raise_error_key_missing: if False, will not raise an error if there are keys that do not exist in some of the samples. Instead will set those values to None.
        """
        super().__init__(skip_keys, raise_error_key_missing)
        self._special_handlers_keys = {}
        if special_handlers_keys is not None:
            self._special_handlers_keys.update(special_handlers_keys)
        self._special_handlers_keys[get_sample_id_key()] = CollateDefault.just_collect_to_list

    def __call__(self, samples: List[Dict]) -> Dict:
        """
        collate list of samples into batch_dict
        :param samples: list of samples
        :return: batch_dict
        """
        batch_dict = NDict()

        # collect all keys
        keys = self._collect_all_keys(samples)

        # collect values
        for key in keys:

            # skip keys
            if key in self._skip_keys:
                continue

            try:
                # collect values into a list
                collected_values, has_error = self._collect_values_to_list(samples, key)

                # batch values
                self._batch_dispatch(batch_dict, samples, key, has_error, collected_values)
            except:
                print(f"Error: Failed to collect key {key}")
                raise

        return batch_dict

    def _batch_dispatch(
        self, batch_dict: dict, samples: List[dict], key: str, has_error: bool, collected_values: list
    ) -> None:
        """
        dispatch a key into collate function and save it into batch_dict
        :param batch_dict: batch dictionary to update
        :param samples: list of samples
        :param key: key to collate
        :param has_error: True, if the key is missing in one of the samples
        :param collected values: the values collected from samples
        :return: nothing - the new batch will be added to batch_dict
        """
        if has_error:
            # do nothing when error occurs
            batch_dict[key] = collected_values
        elif key in self._special_handlers_keys:
            # use special handler if specified
            batch_dict[key] = self._special_handlers_keys[key](collected_values)
        elif isinstance(collected_values[0], (torch.Tensor, np.ndarray, float, int, str, bytes)):
            # batch with default PyTorch implementation
            batch_dict[key] = default_collate(collected_values)
        else:
            batch_dict[key] = collected_values

    @staticmethod
    def just_collect_to_list(values: List[Any]):
        """
        special handler doing nothing - will just keep the collected list
        """
        return values

    @staticmethod
    def pad_all_tensors_to_same_size(values: List[torch.Tensor], pad_val: float = 0.0):
        """
        pad tensors and create a batch - the shape will be the max size per dim
        values: list of tensor - all should have the same number of dimensions
        pad_val: constant value for padding
        :return: torch.stack of padded tensors
        """

        # verify all are tensor and that they have the same dim size
        assert isinstance(values[0], torch.Tensor), f"Expecting just tensors, got {type(values[0])}"
        num_dims = len(values[0].shape)
        for value in values:
            assert isinstance(value, torch.Tensor), f"Expecting just tensors, got {type(value)}"
            assert (
                len(value.shape) == num_dims
            ), f"Expecting all tensors to have the same dim size, got {len(value.shape)} and {num_dims}"

        # get max per dim
        max_per_dim = np.amax(np.stack([value.shape for value in values]), axis=0)

        # pad
        def _pad_size(value, dim):
            assert max_per_dim[dim] >= value.shape[dim]
            return [0, max_per_dim[dim] - value.shape[dim]]

        padded_values = []

        for value in values:
            padding = []
            # F.pad padding description is expected to be provided in REVERSE order (see torch.nn.functional.pad doc)
            for dim in reversed(range(num_dims)):
                padding += _pad_size(value, dim)
            padded_value = F.pad(value, padding, mode="constant", value=pad_val)
            padded_values.append(padded_value)

        return default_collate(padded_values)



import inspect
from typing import Callable, Any, Type, Optional, Sequence
from inspect import stack
import warnings


def get_function_call_str(func, *_args, **_kwargs) -> str:
    """
    Converts a function and its kwargs into a hash value which can be used for caching.
    NOTE:
    1. This is far from being bulletproof, the op might call another function which is not covered and is changed,
    which will make the caching processing be unaware.
    2. This is a mechanism that helps to spot SOME of such issues, NOT ALL
    3. Only a specific subset of arg types contribute to the caching, mainly simple native python types.
    see 'value_to_string' for more details.
    For example, if an arg is an entire numpy array, it will not contribute to the total hash.
    The reason is that it will make the cache calculation too slow, and might
    """

    kwargs = convert_func_call_into_kwargs_only(func, *_args, **_kwargs)

    args_flat_str = func.__name__ + "@"
    args_flat_str += "@".join(["{}@{}".format(str(k), value_to_string(kwargs[k])) for k in sorted(kwargs.keys())])
    args_flat_str += "@" + str(
        inspect.getmodule(func)
    )  # adding full (including scope) name of the function, for the case of multiple functions with the same name
    args_flat_str += "@" + inspect.getsource(func)  # considering the source code (first level of it...)

    return args_flat_str


def value_to_string(val: Any, warn_on_types: Optional[Sequence] = None) -> str:
    """
    Used by default in several caching related hash builders.
    Ignores <...> string as they usually change between different runs
    (for example, due to pointing to a specific memory address)
    """
    if warn_on_types is not None:
        if isinstance(val, tuple(list(warn_on_types))):
            warnings.warn(
                f"type {type(val)} is possibly participating in hashing, this is usually not optimal performance wise."
            )
    ans = str(val)
    if ans.startswith("<"):
        return ""
    return str(val)


def convert_func_call_into_kwargs_only(func: Callable, *args, **kwargs) -> dict:
    """
    considers positional and kwargs (including their default values !)
    and converts into ONLY kwargs
    """
    signature = inspect.signature(func)

    my_kwargs = {k: v.default for k, v in signature.parameters.items() if v.default is not inspect.Parameter.empty}

    # convert positional args into kwargs
    # uses the fact that zip stops on the smallest length ( so only as much as len(args))
    for curr_pos_arg, pos_arg_name in zip(args, inspect.getfullargspec(func).args):
        my_kwargs[pos_arg_name] = curr_pos_arg

    my_kwargs.update(kwargs)

    return my_kwargs


def get_callers_string_description(
    max_look_up: int,
    expected_class: Type,
    expected_function_name: str,
    value_to_string_func: Callable = value_to_string,
    ignore_first_frames=3,  # one for this function, one for HashableCallable, and one for OpBase
):
    """
    iterates on the callstack, and accumulates a string representation of the callers args.
    Used in OpBase to "record" the __init__ args, to be used in the string representation of an Op,
    which is used for building a hash value for samples caching in SamplesCacher

    example call:

    class A:
        def __init__(self):
            text = get_callers_string_description(4, A,

    class B(A):
        def __init__(self, blah, blah2):
            super().__init__()
            #... some logic



    :param max_look_up: how many stack frames to look up
    :param expected_class: what class is the method expected to be,
        stack frames in a different class will be skipped.
        pass None for not requiring any class
    :param expected_function_name: what is the name of the function to allow,
        stack frames in a different function name will be skipped,
        pass None for not requiring anything
    :param value_to_string_func: allows to provide a custom function for converting values to strings
    :param
    """

    str_desc = ""
    try:
        curr_stack = stack()
        curr_locals = None
        # note: frame 0 is this function, frame 1 is whoever called this (and wanted to know about its callers),
        # so both frames 0+1 are skipped.
        for i in range(ignore_first_frames, min(len(curr_stack), max_look_up + ignore_first_frames)):
            curr_locals = curr_stack[i].frame.f_locals
            if expected_class is not None:
                if "self" not in curr_locals:
                    continue
                if not isinstance(curr_locals["self"], expected_class):
                    continue

            if expected_function_name is not None:
                if expected_function_name != str(curr_stack[i].function):
                    continue

            curr_str = ".".join(
                [
                    str(curr_locals["self"].__module__),  # module is probably not needed as class already contains it
                    str(curr_locals["self"].__class__),
                    str(curr_stack[i].function),
                ]
            )

            curr_str += inspect.getsource(curr_stack[i].frame)
            for k, d in curr_stack[i].frame.f_locals.items():
                if "self" == k:
                    continue
                if k.startswith("__"):
                    continue
                curr_str += "@" + str(k) + "@" + value_to_string_func(d)

            str_desc += curr_str

    finally:
        del curr_locals
        del curr_stack

    return str_desc



from functools import partial
from typing import Any, Callable, Dict, Hashable, Optional, Sequence
from fuse.eval.metrics.metrics_common import MetricWithCollectorBase
from fuseimg.data.ops.ops_debug import OpVis2DImage


class MetricVis2DImage(MetricWithCollectorBase):
    """
    Visualize a 2D image (either display it or save it as an image).
    Supports both rgb and gray scale and supports few image formats "channels_first"/"no_channels"/"channels_last"
    """

    def __init__(
        self,
        sample_ids: Optional[Sequence[Hashable]] = None,
        num_samples: Optional[int] = None,
        show: bool = True,
        path: str = ".",
        image_format: str = "channels_first",
        image_process_func: Optional[Callable] = None,
        **kwargs: dict,
    ) -> None:
        """
        :param sample_ids: apply for the specified sample ids. To apply for all set to None.
        :param num_samples: apply for the first num_samples (per process). if None, will apply for all.
        :param show: if set to true will display it, otherwise will save it in `path`
        :param path: location to save the images. Used when `show` set to False
        :param image_format: the format of the image: "channels_first"/"no_channels"/"channels_last"
        :param image_processing_func: callable that get the image (numpy array) and can process it. Typically use to move the range to values matplotlib can display.
        :param kwargs: See OpVis2DImage.call_debug() arguments
        """
        self._op_vis = OpVis2DImage(
            sample_ids=sample_ids,
            num_samples=num_samples,
            show=show,
            path=path,
            image_format=image_format,
            image_process_func=image_process_func,
        )

        op_vis_func = partial(self._op_vis, **kwargs)
        super().__init__(pre_collect_process_func=op_vis_func)
        self._epoch = -1

    def reset(self) -> None:
        self._epoch += 1
        super().reset()
        self._op_vis.reset(f"epoch_{self._epoch}")

    def eval(self, results: Dict[str, Any] = None, ids: Optional[Sequence[Hashable]] = None) -> None:
        # Do nothing - the logic run pre sample in pre_collect_process func
        pass



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import sys
import os
import os.path as path
import fileinput

src_extensions = [".py"]
preamble_signature = "(C) Copyright 2021 IBM Corp."
preamble = """\"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

\"""

"""


def is_preamble_required_file(f):
    if f == "__init__.py":
        return False
    results = [f.endswith(ext) for ext in src_extensions]
    return True in results


def is_header_missing(f):
    with open(f) as reader:
        lines = reader.read().lstrip().splitlines()
        if len(lines) > 1:
            return not lines[1].startswith(preamble_signature)
        return True


def get_src_files(dirname):
    src_files = []
    for cur, _dirs, files in os.walk(dirname):
        [src_files.append(path.join(cur, f)) for f in files if is_preamble_required_file(f)]

    return [f for f in src_files if is_header_missing(f)]


def add_preamble(files, header):
    for line in fileinput.input(files, inplace=True):
        if fileinput.isfirstline():
            [print(h) for h in header.splitlines()]
        print(line, end="")


if __name__ == "__main__":
    """
    Add preamble to .py files except __init__.py.
    The script scans for source files recursively from <root dir> and list the files that are missing the preamble.
    Then, it prompts for confirmation, and if 'y', the modification is made inplace.

    usage:
      add_preamble.py <root dir>
    """

    if len(sys.argv) < 2:
        print("usage: %s <root dir>" % sys.argv[0])
        exit()

    args = [path.abspath(arg) for arg in sys.argv]
    root_path = path.abspath(args[1])

    files = get_src_files(root_path)

    print("Files with missing preamble:")
    [print("  - %s" % f) for f in files]

    if len(files) > 0:
        confirm = input("proceed ? [y/N] ")
        if confirm != "y":
            exit(0)

        add_preamble(files, preamble)
    else:
        print("None")



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import torch
from typing import Sequence, Dict, Tuple, Callable, Optional

from fuse.utils.ndict import NDict


class ModelMultistream(torch.nn.Module):
    """
    Multi-stream Fuse model - convolutional neural network with multiple processing streams and multiple heads
    """

    def __init__(
        self,
        conv_inputs: Tuple[str, int],
        backbone_streams: Sequence[torch.nn.Module],
        heads: Sequence[torch.nn.Module],
        split_logic: Optional[Callable] = None,
        join_logic: Optional[Callable] = None,
    ) -> None:
        """
        Multi-stream Fuse model - convolutional neural network with multiple processing streams and multiple heads
        :param conv_inputs:             batch_dict name for model input and its number of input channels
            for example: conv_inputs=('data.input.input_0.tensor', 1)
        :param backbone_streams:        List of PyTorch backbone modules - one per stream (can share weights or not)
        :param heads:                   Sequence of head modules
            for example: heads=(HeadGlobalPoolingClassifier(conv_inputs = (('model.backbone_features', 384),))
        :param split_logic:             Optional callable, splits input into streams. If None, sends each input channel to consecutive stream.
                                            Signature: stream_outputs = split_logic(batch_dict, backbone_streams)
        :param join_logic:              Optional callable, joins stream outputs into single feature map. If None, concatenates on channel axis.
                                            Signature: feature_map = join_logic(batch_dict, stream_outputs)
        """
        super().__init__()
        assert (
            conv_inputs is not None
        ), "You must provide conv_inputs - for example: conv_inputs=('data.input.input_0.tensor', 1)"
        self.conv_inputs = conv_inputs
        self.split_logic = split_logic
        self.join_logic = join_logic

        # Register modules
        self.backbone_streams = torch.nn.ModuleList(backbone_streams)
        self.add_module("backbones", self.backbone_streams)
        assert (
            heads is not None
        ), "You must provide heads - for example: heads=(HeadGlobalPoolingClassifier(conv_inputs = (('model.backbone_features', 384),)),)"
        self.heads = torch.nn.ModuleList(heads)
        self.add_module("heads", self.heads)

    def forward(self, batch_dict: NDict) -> Dict:

        # Forward pass through multiple streams
        # -------------------------------------
        if self.split_logic is None:
            # If no split logic is provided, send each channel to different stream
            conv_input = batch_dict[self.conv_inputs[0]]  # shape = [batch_size, num_channels, height, width]
            stream_outputs = []
            for ch_idx in range(conv_input.shape[1]):
                single_channel_batch = conv_input[:, ch_idx, :, :].unsqueeze(
                    dim=1
                )  # shape = [batch_size, 1, height, width]
                stream_output = self.backbone_streams[ch_idx](single_channel_batch)
                stream_outputs.append(stream_output)
        elif callable(self.split_logic):
            stream_outputs = self.split_logic(batch_dict, self.backbone_streams)
        else:
            raise Exception("Error in ModelMultistream - bad split logic provided")

        # Combining feature maps from multiple streams
        # --------------------------------------------
        if self.join_logic is None:
            # If no join logic is provided, concatenate feature maps in channel axis
            backbone_features = torch.cat(stream_outputs, dim=1)
        elif callable(self.join_logic):
            backbone_features = self.join_logic(batch_dict, stream_outputs)
        else:
            raise Exception("Error in ModelMultistream - bad join logic provided")

        batch_dict["model.backbone_features"] = backbone_features
        for head in self.heads:
            batch_dict = head.forward(batch_dict)

        return batch_dict



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""
from abc import abstractmethod
from typing import Any, List, Optional, Sequence, Union
from fuse.data.ops.op_base import OpReversibleBase
import numpy as np

from fuse.data import OpBase
import torch
from torch import Tensor
from fuse.utils.ndict import NDict


class Cast:
    """
    Cast methods
    """

    @staticmethod
    def to_tensor(value: Any, dtype: Optional[torch.dtype] = None, device: Optional[torch.device] = None) -> Tensor:
        """
        Convert many types to tensor
        """
        if isinstance(value, torch.Tensor) and dtype is None and device is None:
            pass  # do nothing
        elif isinstance(value, (torch.Tensor)):
            value = value.to(dtype=dtype, device=device)
        elif isinstance(value, (np.ndarray, int, float, list)):
            value = torch.tensor(value, dtype=dtype, device=device)
        else:
            raise Exception(f"Unsupported type {type(value)} - add here support for this type")

        return value

    @staticmethod
    def to_numpy(value: Any, dtype: Optional[np.dtype] = None) -> np.ndarray:
        """
        Convert many types to numpy
        """
        if isinstance(value, np.ndarray) and dtype is None:
            pass  # do nothing
        elif isinstance(value, (torch.Tensor, int, float, list, np.ndarray)):
            value = np.array(value, dtype=dtype)
        else:
            raise Exception(f"Unsupported type {type(value)} - add here support for this type")

        return value

    @staticmethod
    def to_int(value: Any) -> np.ndarray:
        """
        Convert many types to int
        """
        if isinstance(value, int):
            pass  # do nothing
        elif isinstance(value, (torch.Tensor, np.ndarray, float, str, bytes)):
            value = int(value)
        else:
            raise Exception(f"Unsupported type {type(value)} - add here support for this type")

        return value

    @staticmethod
    def to_float(value: Any) -> np.ndarray:
        """
        Convert many types to float
        """

        if isinstance(value, float):
            pass  # do nothing
        elif isinstance(value, (torch.Tensor, np.ndarray, int, str)):
            value = float(value)
        else:
            raise Exception(f"Unsupported type {type(value)} - add here support for this type")

        return value

    @staticmethod
    def to_list(value: Any) -> np.ndarray:
        """
        Convert many types to list
        """

        if isinstance(value, list):
            pass  # do nothing
        elif isinstance(value, (torch.Tensor, np.ndarray)):
            value = value.tolist()
        else:
            raise Exception(f"Unsupported type {type(value)} - add here support for this type")

        return value

    @staticmethod
    def to(value: Any, type_name: str, **kwargs) -> Any:
        """
        Convert any type to type specified in type_name
        """

        if type_name == "ndarray":
            return Cast.to_numpy(value, **kwargs)
        if type_name == "Tensor":
            return Cast.to_tensor(value, **kwargs)
        if type_name == "float":
            return Cast.to_float(value, **kwargs)
        if type_name == "int":
            return Cast.to_int(value, **kwargs)
        if type_name == "list":
            return Cast.to_list(value, **kwargs)

    @staticmethod
    def like(value: Any, like_value: Any) -> Any:
        """
        Convert any type to type specified in type_name
        """

        if isinstance(like_value, np.ndarray):
            return Cast.to_numpy(value, dtype=like_value.dtype)
        if isinstance(like_value, torch.Tensor):
            return Cast.to_tensor(value, dtype=like_value.dtype, device=like_value.device)
        if isinstance(like_value, float):
            return Cast.to_float(value)
        if isinstance(like_value, int):
            return Cast.to_int(value)
        if isinstance(like_value, list):
            return Cast.to_list(value)


class OpCast(OpReversibleBase):
    def __call__(
        self, sample_dict: NDict, op_id: Optional[str], key: Union[str, Sequence[str]], **kwargs
    ) -> Union[None, dict, List[dict]]:
        """
        See super class
        :param key: single key or list of keys from sample_dict to convert
        """
        if isinstance(key, str):
            keys = [key]
        else:
            keys = key

        for key_name in keys:
            value = sample_dict[key_name]
            sample_dict[f"{op_id}_{key_name}"] = type(value).__name__
            value = self._cast(value, **kwargs)
            sample_dict[key_name] = value

        return sample_dict

    def reverse(self, sample_dict: NDict, key_to_reverse: str, key_to_follow: str, op_id: Optional[str]) -> dict:
        type_name = sample_dict[f"{op_id}_{key_to_follow}"]
        value = sample_dict[key_to_reverse]
        value = Cast.to(value, type_name)
        sample_dict[key_to_reverse] = value

        return sample_dict

    @abstractmethod
    def _cast(self):
        raise NotImplementedError


class OpToTensor(OpCast):
    """
    Convert many types to tensor
    """

    def _cast(self, value: Any, dtype: Optional[torch.dtype] = None, device: Optional[torch.device] = None) -> Tensor:
        return Cast.to_tensor(value, dtype, device)


class OpToNumpy(OpCast):
    """
    Convert many types to numpy
    """

    def _cast(self, value: Any, dtype: Optional[np.dtype] = None) -> np.ndarray:

        return Cast.to_numpy(value, dtype)


class OpToInt(OpCast):
    """
    Convert many types to int
    """

    def _cast(self, value: Any) -> int:
        return Cast.to_int(value)


class OpToFloat(OpCast):
    """
    Convert many types to float
    """

    def _cast(self, value: Any) -> float:
        return Cast.to_float(value)


class OpOneHotToNumber(OpBase):
    """
    Convert one-hot encoding into numbers.
    for example:
        [1, 0, 0, 0] -> 0
        [0, 0, 1, 0] -> 2
    """

    def __init__(self, num_classes, verify_arguments: bool = True):
        """
        :param num_classes: Number of class the Op should expect.
        :param verify_arguments: Defualt is True - can be set to False for a speedup.
        """
        super().__init__()
        self._num_classes = num_classes
        self._verify_arguments = verify_arguments

    def __call__(self, sample_dict: NDict, key: str, **kwargs) -> Union[None, dict, List[dict]]:
        """
        :param key: the sample_dict's key where the one-hot vector is located.
                    The corresponding number will be save in the same key (instead of the one-hot)
        """
        one_hot_vector = np.array(sample_dict[key])

        if self._verify_arguments:
            assert self._num_classes == len(
                one_hot_vector
            ), f"Error: One-hot vector length is {len(one_hot_vector)} but number of classes is {self._num_classes}"
            assert self.is_one_hot_vector(one_hot_vector), f"expect one-hot vector, instead got: {one_hot_vector} ."

        value = one_hot_vector.argmax()
        sample_dict[key] = torch.tensor(value)

        return sample_dict

    def is_one_hot_vector(self, vector: np.ndarray) -> bool:
        """
        Return true iff vector is one-hot with unique class. (not multiple features)

        :param vector:
        """

        max_value = vector.max()
        min_value = vector.min()
        sum_value = vector.sum()

        ans = (max_value == 1) and (min_value == 0) and (sum_value == 1)
        return ans



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import List, Optional, Union, Sequence
from fuse.data.datasets.caching.samples_cacher import SamplesCacher
from fuse.data.utils.sample import get_sample_id

from torch.utils.data import Dataset

from fuse.data.ops.op_base import OpBase
from fuse.data.datasets.dataset_default import DatasetDefault
from fuse.data.pipelines.pipeline_default import PipelineDefault
from fuse.utils.ndict import NDict


# Dataset processor
class OpReadDataset(OpBase):
    """
    Op that extract data from pytorch dataset that returning sequence of values and adds those values to sample_dict
    """

    def __init__(self, dataset: Dataset, sample_keys: Sequence[str]):
        """
        :param dataset: the pytorch dataset to convert. The dataset[i] expected to return sequence of values or a single value
        :param sample_keys: sequence keys - naming each value returned by dataset[i]
        """
        # store input arguments
        super().__init__()
        self._sample_keys = sample_keys
        self._dataset = dataset

    def __call__(self, sample_dict: NDict) -> Union[None, dict, List[dict]]:
        """
        See super class
        """
        # extact dataset index
        name, dataset_index = get_sample_id(sample_dict)

        # extract values
        sample_values = self._dataset[dataset_index]
        if not isinstance(sample_values, Sequence):
            sample_values = [sample_values]
        assert len(self._sample_keys) == len(
            sample_values
        ), f"Error: expecting dataset[i] to return {len(self._sample_keys)} to match sample keys"

        # add values to sample_dict
        for key, elem in zip(self._sample_keys, sample_values):
            sample_dict[key] = elem
        return sample_dict


class DatasetWrapSeqToDict(DatasetDefault):
    """
    Fuse Dataset Wrapper
    wraps pytorch sequence dataset (pytorch dataset in which each sample, dataset[i] is a sequence of values).
    Each value extracted from pytorch sequence dataset will be added to sample_dict.
    Plus this dataset inherits all DatasetDefault features

    Example:
        torch_seq_dataset = torchvision.datasets.MNIST(path, download=True, train=True)
        # wrapping torch dataset
        dataset = DatasetWrapSeqToDict(name='train', dataset=torch_seq_dataset, sample_keys=('data.image', 'data.label'))
        train_dataset.create()

        # get sample
        sample = train_dataset[index] # sample is a dict with keys: 'data.sample_id', 'data.image' and 'data.label'
    """

    def __init__(
        self,
        name: str,
        dataset: Dataset,
        sample_keys: Union[Sequence[str], str],
        cache_dir: Optional[str] = None,
        **kwargs,
    ):
        """
        :param name: name of the data extracted from dataset, typically: 'train', 'validation;, 'test'
        :param dataset: the dataset to extract the data from
        :param sample_keys: sequence keys - naming each value returned by dataset[i]
        :param cache_dir: Optional - provied a path in case caching is required to help optimize the running time
        :param kwargs: optional, additional arguments to provide to DatasetDefault
        """
        sample_ids = [(name, i) for i in range(len(dataset))]
        static_pipeline = PipelineDefault(name="staticp", ops_and_kwargs=[(OpReadDataset(dataset, sample_keys), {})])
        if cache_dir is not None:
            cacher = SamplesCacher("dataset_test_cache", static_pipeline, cache_dir, restart_cache=True)
        else:
            cacher = None
        super().__init__(sample_ids=sample_ids, static_pipeline=static_pipeline, cacher=cacher, **kwargs)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Dict, Tuple, Sequence, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

from fuse.dl.models.heads.common import ClassifierFCN, ClassifierMLP
from fuse.utils.ndict import NDict


class HeadGlobalPoolingClassifier(nn.Module):
    def __init__(
        self,
        head_name: str = "head_0",
        conv_inputs: Sequence[Tuple[str, int]] = None,
        num_classes: int = 2,
        tabular_data_inputs: Optional[Sequence[Tuple[str, int]]] = None,
        pooling: str = "max",
        layers_description: Sequence[int] = (256,),
        tabular_layers_description: Sequence[int] = tuple(),
        dropout_rate: float = 0.1,
        tabular_dropout_rate: float = 0.0,
        shared_classifier_head: Optional[torch.nn.Module] = None,
    ) -> None:
        """
        Classifier head with a global pooling operator.

        Output of a forward pass:
        'model.logits.head_name' and 'outputs.head_name', both in shape [batch_size, num_classes]

        :param head_name:                   batch_dict key
        :param conv_inputs:                 List of feature map inputs - tuples of (batch_dict key, channel depth)
                                            If multiple inputs are used, they are concatenated on the channel axis
                example: conv_inputs = (('model.backbone_features', 384),)
        :param num_classes:                 Number of output classes (per feature map location)
        :param tabular_data_inputs:          Additional vector (one dimensional) inputs, concatenated just before the classifier module
        :param pooling:                     Type of global pooling operator ('max' or 'avg')
        :param layers_description:          Layers description for the classifier module - sequence of hidden layers sizes
        :param tabular_layers_description: Layers description for the tabular data, before the concatination with the features extracted from the image - sequence of hidden layers sizes
        :param dropout_rate:                Dropout rate for classifier module layers
        :param tabular_dropout_rate:        Dropout rate for tabular layers
        :param shared_classifier_head:      Optional reference for external torch.nn.Module classifier
        """
        super().__init__()

        assert pooling in ("max", "avg")

        self.head_name = head_name
        assert conv_inputs is not None, "conv_inputs must be provided"
        self.conv_inputs = conv_inputs
        self.tabular_data_inputs = tabular_data_inputs
        self.pooling = pooling

        feature_depth = sum([conv_input[1] for conv_input in self.conv_inputs])

        if tabular_data_inputs is not None:
            if len(tabular_layers_description) == 0:
                feature_depth += sum([post_concat_input[1] for post_concat_input in tabular_data_inputs])
                self.tabular_module = nn.Identity()
            else:
                feature_depth += tabular_layers_description[-1]
                self.tabular_module = ClassifierMLP(
                    in_ch=sum([post_concat_input[1] for post_concat_input in tabular_data_inputs]),
                    num_classes=None,
                    layers_description=tabular_layers_description,
                    dropout_rate=tabular_dropout_rate,
                )

        if shared_classifier_head is not None:
            self.classifier_head_module = shared_classifier_head
        else:
            self.classifier_head_module = ClassifierFCN(
                in_ch=feature_depth,
                num_classes=num_classes,
                layers_description=layers_description,
                dropout_rate=dropout_rate,
            )

    def forward(self, batch_dict: NDict) -> Dict:
        conv_input = torch.cat([batch_dict[conv_input[0]] for conv_input in self.conv_inputs])

        if len(conv_input.shape) == 2:
            res = conv_input
        else:
            # Global pooling - input shape is [batch_size, num_channels, height, width], result is [batch_size, num_channels, 1, 1]
            if self.pooling == "max":
                res = F.max_pool2d(conv_input, kernel_size=conv_input.shape[2:])
            elif self.pooling == "avg":
                res = F.avg_pool2d(conv_input, kernel_size=conv_input.shape[2:])

        if self.tabular_data_inputs is not None:
            tabular_input = torch.cat([batch_dict[tabular_input[0]] for tabular_input in self.tabular_data_inputs])
            tabular_input = self.tabular_module(tabular_input)
            tabular_input = tabular_input.reshape(tabular_input.shape + (1, 1))
            res = torch.cat([res, tabular_input], dim=1)

        logits = self.classifier_head_module(res)  # --> res.shape = [batch_size, 2, 1, 1]
        if len(logits.shape) > 2:
            logits = logits.squeeze(dim=3)  # --> res.shape = [batch_size, 2, 1]
            logits = logits.squeeze(dim=2)  # --> res.shape = [batch_size, 2]

        cls_preds = F.softmax(logits, dim=1)

        batch_dict["model.logits." + self.head_name] = logits
        batch_dict["model.output." + self.head_name] = cls_preds

        return batch_dict



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""
import logging
import os

from typing import Union
import pandas as pd

from fuse.utils.utils_logger import fuse_logger_start
from fuse.utils.file_io.file_io import save_dataframe


def make_targets_file(data_path: str, split: Union[str, dict], output_filename: str):
    """
    Automaitically make targets file in the requested format
    :param data_path: path to the original data downloaded from https://github.com/neheller/KNIGHT
    :param cache_path: Optional - path to the cache folder. If none, it will pre-processes the data again
    :param split: either path to pickled dictionary or the actual dictionary specifing the split between train and validation. the dictionary maps "train" to list of sample descriptors and "val" to list of sample descriptions
    :param output_filename: filename of the output csv file
    """
    # Logger
    fuse_logger_start(console_verbose_level=logging.INFO)
    lgr = logging.getLogger("Fuse")
    lgr.info("KNIGHT: make targets file", {"attrs": ["bold", "underline"]})
    lgr.info(f"targets_filename={os.path.abspath(output_filename)}", {"color": "magenta"})

    # Data
    # read train/val splits file.
    # use validation set if split specified, otherwise assume testset
    if isinstance(split, str):
        is_validation_set = True
    else:
        assert split is None, f"Error: unexpected split format {split}"
        is_validation_set = False

    if is_validation_set:
        split = pd.read_pickle(split)
        if isinstance(split, list):
            # For this example, we use split 0 out of the the available cross validation splits
            split = split[0]

            json_labels_filepath = os.path.join(data_path, "knight.json")
            labels = pd.read_json(json_labels_filepath)
            labels = pd.DataFrame({"target": labels["aua_risk_group"].values})

    else:  # test mode
        # if this function is ran in test mode, then presumably the user has the test labels
        # for test purposes, the test labels file isn't shared with participants:
        json_labels_filepath = os.path.join(data_path, "knight_test_labels.json")
        if not os.path.isfile(json_labels_filepath):
            ValueError("No test labels file found")
        labels = pd.read_json(json_labels_filepath, typ="series")
        labels = pd.DataFrame({"target": labels.values})

    levels = ["benign", "low_risk", "intermediate_risk", "high_risk", "very_high_risk"]
    labels["Task2-target"] = labels["target"].apply(lambda x: levels.index(x))
    labels.insert(1, "Task1-target", labels["Task2-target"].apply(lambda x: int(x > 3)))
    labels.reset_index(inplace=True)
    labels.rename({"index": "case_id"}, axis=1, inplace=True)
    labels.drop(["target"], axis=1, inplace=True)
    save_dataframe(labels, output_filename, index=False)
    return


if __name__ == "__main__":
    """
    Automaitically make targets file (csv files that holds just the labels for tasks) in the requested format
    Usage: python make_predictions_file <data_path> <cache_path> <split> <output_filename>
    See details in function make_predictions_file.
    """
    # no arguments - set arguments inline - see details in function make_targets_file
    data_path = ""
    split = "baseline/splits_final.pkl"
    output_filename = "validation_targets.csv"
    make_targets_file(data_path=data_path, split=split, output_filename=output_filename)



from typing import Optional
from time import time
from deepdiff import DeepDiff
from fuse.data import get_sample_id

"""
By auditing the samples, "stale" caches can be found, which is very important to detect.
A stale cache of a sample is a cached sample which contains different information then the same sample as it is being freshly created.
There are several reasons that it can happen, for example, a change in some code dependency in some operation in the sample processing pipeline.
Note - setting a too high audit frequency will slow your training.
audit example usage:
# a minimalistic approach, testing only the first sample. Almost no slow down of entire train session, but not periodic audit so higher chance to miss a stale cached sample.
SampleCachingAudit(audit_first_sample=True,audit_rate=None)
)

#another audit usage example - in this case the first sample will be audited, and also one sample every 20 minutes
SampleCachingAudit(audit_first_sample=True, audit_rate=20, audit_units='minutes')
)
"""


class SampleCachingAudit:
    def __init__(
        self,
        audit_first_sample: bool = True,
        audit_rate: Optional[int] = 30,
        audit_units: str = "minutes",
        **audit_diff_kwargs: Optional[dict],
    ):
        """
        :param audit_rate: how frequently, a sample will be both loaded from cache AND loaded fully without using cache.
        Pass 0 or None to disable.
        The purpose of this is to detect cases in which the cached samples no longer match the sample loading sequence of Ops,
        and a cache reset is required.
        Will be ignored if no cacher is provided.
        :param audit_units: the units in which audit_rate will be used. Supported options are ['minutes', 'samples']
        Will be ignored if no cacher is provided.
        :param **audit_diff_kwargs: optionally, pass custom kwargs to DeepDiff comparison.
        This is useful if, for example, you want small epsilon differences to be ignored.
        In such case, you can provide math_epsilon=1e-9 to avoid throwing exception for small differences

        See their documentation here to learn about additional possible flags: https://zepworks.com/deepdiff/current/
        Important - as a default, we pass ignore_nan_inequality=True, as we think it's the most reasonable comparison strategy suitable for caches.
        You can use ignore_nan_inequality=False if you prefer the original default behavior (which considers NaN to not be equal NaN)

        """

        if "ignore_nan_inequality" not in audit_diff_kwargs:
            audit_diff_kwargs["ignore_nan_inequality"] = True
        _audit_unit_options = ["minutes", "samples", None]
        if audit_units not in _audit_unit_options:
            raise Exception(f"audit_units must be one of {_audit_unit_options}")
        self._audit_rate = audit_rate
        self._audit_first_sample = audit_first_sample
        self._audited_so_far = 0
        if self._audit_rate == 0:
            self._audit_rate = None
        self._audit_units = audit_units
        self._audit_units_passed_since_last_audit = 0.0
        if self._audit_units == "minutes":
            self._prev_time = time()
        self._audit_diff_kwargs = audit_diff_kwargs

    def update(self) -> bool:
        """
        Updates internal state related to the audit features (comparison of a sample loaded from cache with a fully loaded/processed sample)
        returns whether an audit should occur now or not.
        """
        if (self._audit_first_sample) and (self._audited_so_far == 0):
            return True
        if self._audit_rate is not None:
            # progress audit units passed so far
            if self._audit_units == "minutes":
                self._audit_units_passed_since_last_audit += (time() - self._prev_time) / 60.0
                self._prev_time = time()
            elif self._audit_units == "samples":
                self._audit_units_passed_since_last_audit += 1
            else:
                assert False

            # check if we need an audit now
            if self._audit_units_passed_since_last_audit >= self._audit_rate:
                # reset it
                if self._audit_units == "minutes":
                    self._audit_units_passed_since_last_audit %= self._audit_rate
                else:
                    self._audit_units_passed_since_last_audit = 0.0
                return True
        return False

    def audit(self, cached_sample, fresh_sample):
        diff = DeepDiff(cached_sample, fresh_sample, **self._audit_diff_kwargs)
        self._audited_so_far += 1
        if len(diff) > 0:
            raise Exception(
                f"Error! During AUDIT found a mismatch between cached_sample and loaded sample.\n"
                "Please reset your cache.\n"
                "Note - this can happen if a change in your (static) pipeline Ops is not expressed in the calculated hash function.\n"
                "There are several reasons that can cause this, for example, you are calling, from within your op external code.\n"
                "This is perfectly fine to do, just make sure you reset your cache after such change.\n"
                "Gladly, the Audit feature caught this stale cache state! :)\n"
                f"sample id in which this staleness was caught: {get_sample_id(fresh_sample)}\n"
                'NOTE: if small changes between the saved cached and the live-loaded/processed sample are ok for your use case, you can set a tolerance epsilon like this: audit_diff_kwargs={"math_epsilon":1e-9}'
            )



#!/usr/bin/env python
import os
import pathlib
from setuptools import setup, find_namespace_packages


# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
with open(os.path.join(HERE, "README.md"), "r", encoding="utf-8") as fh:
    long_description = fh.read()

# list of requirements for core packages
fuse_requirements = []
with open(os.path.join(HERE, "fuse/requirements.txt"), "r") as fh:
    for line in fh:
        if not line.startswith("#"):
            fuse_requirements.append(line.strip())

# list of requirements for core packages for development
fuse_requirements_dev = []
with open(os.path.join(HERE, "fuse/requirements_dev.txt"), "r") as fh:
    for line in fh:
        if not line.startswith("#"):
            fuse_requirements_dev.append(line.strip())

# list of requirements for fuseimg
fuseimg_requirements = []
with open(os.path.join(HERE, "fuseimg/requirements.txt"), "r") as fh:
    for line in fh:
        if not line.startswith("#"):
            fuseimg_requirements.append(line.strip())

# all extra requires
all_requirements = fuseimg_requirements + fuse_requirements_dev
# version
from fuse.version import __version__  # noqa

version = __version__

setup(
    name="fuse-med-ml",
    version=version,
    description="Open-source PyTorch based framework designed to facilitate deep learning R&D in medical imaging",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/BiomedSciAI/fuse-med-ml/",
    author="IBM Research - Machine Learning for Healthcare and Life Sciences",
    author_email="moshiko.raboh@ibm.com",
    packages=find_namespace_packages(),
    license="Apache License 2.0",
    install_requires=fuse_requirements,
    extras_require={"fuseimg": fuseimg_requirements, "dev": fuse_requirements_dev, "all": all_requirements},
)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import unittest
from fuse.data.ops.op_base import op_call

import pandas as pd
from fuse.utils.ndict import NDict
from fuse.data.ops.ops_read import OpReadDataframe


class TestOpsRead(unittest.TestCase):
    def test_op_read_dataframe(self):
        """
        Test OpReadDataframe
        """
        data = {"sample_id": ["a", "b", "c", "d"], "data.value1": [10, 7, 3, 9], "data.value2": ["5", "4", "3", "2"]}
        df = pd.DataFrame(data)
        op = OpReadDataframe(data=df)
        sample_dict = NDict({"data": {"sample_id": "c"}})
        sample_dict = op_call(op, sample_dict, "id")
        self.assertEqual(sample_dict["data.value1"], 3)
        self.assertEqual(sample_dict["data.value2"], "3")

        op = OpReadDataframe(data=df, columns_to_extract=["sample_id", "data.value2"])
        sample_dict = NDict({"data": {"sample_id": "c"}})
        sample_dict = op_call(op, sample_dict, "id")
        self.assertFalse("data.value1" in sample_dict)
        self.assertEqual(sample_dict["data.value2"], "3")

        op = OpReadDataframe(
            data=df, columns_to_extract=["sample_id", "data.value2"], rename_columns={"data.value2": "data.value3"}
        )
        sample_dict = NDict({"data": {"sample_id": "c"}})
        sample_dict = op_call(op, sample_dict, "id")
        self.assertFalse("data.value1" in sample_dict)
        self.assertFalse("data.value2" in sample_dict)
        self.assertEqual(sample_dict["data.value3"], "3")


if __name__ == "__main__":
    unittest.main()
